import logging
import time
from pathlib import Path

ROOT_PATH = Path(__file__).parent.parent.parent


def logging_config(
        level=logging.INFO,
        log_format: str = '%(asctime)s %(name)-20s %(levelname)-10s %(message)s',
        log_date_fmt: str = '%Y-%m-%dT%H:%M:%S%z',
        **kwargs
):
    """Wrapper for logging basic config with a default format

    :param level: logging level
    :param log_format: logging format
    :param log_date_fmt: logging date format
    :param kwargs:
    :return:
    """
    logging.basicConfig(level=level, format=log_format, datefmt=log_date_fmt, **kwargs)
    logging.Formatter.converter = time.gmtime



"""Aito `Query API <https://aito.ai/docs/api/#query-api>`__ Response Class"""

from abc import ABC, abstractmethod
from typing import TypeVar, Generic, Type, List, Dict, Any

from .aito_response import BaseResponse
from aito.utils._json_format import JsonFormat


class BaseHit(JsonFormat):
    """Aito `ResponseHit <https://aito.ai/docs/api/#schema-response-hit>`__"""
    def __init__(self, json: Dict):
        """

        :param json: the content of the hit
        :type json: Dict
        """
        self.json_schema_validate(json)
        self._json = json

    def __getitem__(self, item):
        if item not in self._json:
            raise KeyError(f'The hit does not contain field `{item}`. '
                           f'Please specify the field in the `select` clause of the query')
        return self._json[item]

    def __contains__(self, item):
        return item in self._json

    def __iter__(self):
        return self._json.keys()

    @classmethod
    def json_schema(cls):
        return {'type': 'object'}

    @property
    def json(self):
        """the content of the hit"""
        return self._json

    def to_json_serializable(self):
        return self._json

    @classmethod
    def from_deserialized_object(cls, obj: Any):
        return cls(obj)


class _BaseScoredHit(BaseHit, ABC):
    """Base class for `ResponseHit <https://aito.ai/docs/api/#schema-response-hit>`__ that contains a
    `score <https://aito.ai/docs/api/#schema-score-field>`__
    """
    @classmethod
    @abstractmethod
    def score_aliases(cls) -> list:
        """The aliases of the score. For example, a predict hit can have both $score and $p
        """
        pass

    @property
    def explanation(self) -> Dict:
        """returns the explanation of how the score was calculated

        :rtype: Dict
        """
        return self.__getitem__('$why')

    @property
    def score(self) -> float:
        """returns the score

        :rtype: float
        """
        score_field = next((alias for alias in self.score_aliases() if alias in self._json), None)
        if score_field is None:
            raise KeyError(f'The hit does not contain the score field. Please specify one of '
                           f'{"|".join(self.score_aliases())} in the `select` clause of the query')
        return self._json[score_field]


class ScoredHit(_BaseScoredHit):
    """`ResponseHit <https://aito.ai/docs/api/#schema-response-hit>`__ that contains a
    `score <https://aito.ai/docs/api/#schema-score-field>`__"""
    @classmethod
    def score_aliases(cls) -> list:
        return ['$score']


class ProbabilityHit(_BaseScoredHit):
    """`ResponseHit <https://aito.ai/docs/api/#schema-response-hit>`__ with probability"""
    @classmethod
    def score_aliases(cls) -> list:
        return ['$score', '$p']

    @property
    def predicted_probability(self) -> float:
        """return the predicted probability

        :rtype: float
        """
        return self.score


class RelateHit(BaseHit):
    """`ResponseHit <https://aito.ai/docs/api/#schema-response-hit>`__ of the
    `Relate query <https://aito.ai/docs/api/#post-api-v1-relate>`__

    Contains statistical information between pair of features
    """
    @property
    def frequencies(self):
        """return frequencies information of the relation"""
        return self.__getitem__('fs')

    @property
    def probabilities(self):
        """return probabilities information of the relation"""
        return self.__getitem__('ps')


HitType = TypeVar('HitType', bound=BaseHit)


class _GenericHitsResponse(BaseResponse, Generic[HitType], ABC):
    """HitsResponse with information of the HitType"""
    def __init__(self, json):
        """

        :param json: the original JSON response
        :type json: Dict
        """
        super().__init__(json)
        self._hits = [self.hit_cls(hit) for hit in json['hits']]

    @property
    @abstractmethod
    def hit_cls(self) -> Type[HitType]:
        """the class of the hit

        :rtype: Type[HitType]
        """
        pass

    @classmethod
    def json_schema(cls):
        return {
            'type': 'object',
            'properties': {
                'offset': {'type': 'integer'},
                'total': {'type': 'integer'},
                'hits': {'type': 'array', 'items': {'type': 'object'}}
            },
            'required': ['offset', 'total', 'hits']
        }

    @property
    def offset(self) -> int:
        """the number of hits that is skipped

        :rtype: int
        """
        return self._json['offset']

    @property
    def total(self) -> int:
        """the total number of hits

        :rtype: int
        """
        return self._json['total']

    @property
    def hits(self) -> List[HitType]:
        """the returned hits

        :rtype: List[HitType]
        """
        return self._hits

    @property
    def first_hit(self) -> HitType:
        """return the first hit

        :rtype: HitType
        """
        return self._hits[0]


class HitsResponse(_GenericHitsResponse[BaseHit]):
    """The response contains entries or `hits <https://aito.ai/docs/api/#schema-hits>`__ returned for a given query"""
    @property
    def hit_cls(self) -> Type[HitType]:
        return BaseHit


class SearchResponse(_GenericHitsResponse[BaseHit]):
    """Response of the `Search query <https://aito.ai/docs/api/#post-api-v1-search>`__"""
    @property
    def hit_cls(self) -> Type[HitType]:
        return BaseHit


class PredictResponse(_GenericHitsResponse[ProbabilityHit]):
    """Response of the `Predict query <https://aito.ai/docs/api/#post-api-v1-predict>`__"""
    @property
    def hit_cls(self) -> Type[HitType]:
        return ProbabilityHit

    @property
    def predictions(self) -> List[ProbabilityHit]:
        """return a list of predictions in descending order of the estimated probability

        :return:
        :rtype:
        """
        # noinspection PyTypeChecker
        return self.hits

    @property
    def top_prediction(self) -> ProbabilityHit:
        """return the prediction with the highest probability

        :rtype: Dict
        """
        # noinspection PyTypeChecker
        return self.first_hit


class RecommendResponse(_GenericHitsResponse[ProbabilityHit]):
    """Response of the `Recommend query <https://aito.ai/docs/api/#post-api-v1-recommend>`__"""
    @property
    def hit_cls(self) -> Type[HitType]:
        return ProbabilityHit

    @property
    def recommendations(self) -> List[ProbabilityHit]:
        """return a list of recommendations in descending order of the estimated probability of the specified goal

        :return:
        :rtype:
        """
        # noinspection PyTypeChecker
        return self.hits

    @property
    def top_recommendation(self) -> ProbabilityHit:
        """return the recommendation with the highest probability of the goal

        :rtype: Dict
        """
        # noinspection PyTypeChecker
        return self.first_hit


class SimilarityResponse(_GenericHitsResponse[ScoredHit]):
    """Response of the `Similarity query <https://aito.ai/docs/api/#post-api-v1-similarity>`__"""
    @property
    def hit_cls(self) -> Type[HitType]:
        return ScoredHit

    @property
    def similar_entries(self) -> List[ScoredHit]:
        """return a list of similar entries in descending order of the similarity score

        :return:
        :rtype:
        """
        # noinspection PyTypeChecker
        return self.hits

    @property
    def most_similar_entry(self) -> ScoredHit:
        """return the most similar entry

        :rtype: Dict
        """
        # noinspection PyTypeChecker
        return self.first_hit


class MatchResponse(_GenericHitsResponse[ProbabilityHit]):
    """Response of the `Match query <https://aito.ai/docs/api/#post-api-v1-match>`__"""
    @property
    def hit_cls(self) -> Type[HitType]:
        return ProbabilityHit

    @property
    def matches(self) -> List[ProbabilityHit]:
        """return a list of matches in descending order of the estimated probability

        :return:
        :rtype:
        """
        # noinspection PyTypeChecker
        return self.hits

    @property
    def top_match(self) -> ProbabilityHit:
        """return the match with the highest probability

        :rtype: Dict
        """
        # noinspection PyTypeChecker
        return self.first_hit


class RelateResponse(_GenericHitsResponse[RelateHit]):
    """Response of the `Relate query <https://aito.ai/docs/api/#post-api-v1-relate>`__"""
    @property
    def hit_cls(self) -> Type[HitType]:
        return RelateHit

    @property
    def relations(self) -> List[RelateHit]:
        # noinspection PyTypeChecker
        return self.hits


class EvaluateResponse(BaseResponse):
    """Response of the `Evaluate query <https://aito.ai/docs/api/#post-api-v1-evaluate>`__"""
    @property
    def accuracy(self) -> float:
        """test evaluation accuracy

        :rtype: float
        """
        return self.__getitem__('accuracy')

    @property
    def test_sample_count(self) -> int:
        """the number of entries in the table that was used as test set

        :rtype: int
        """
        return self.__getitem__('testSamples')

    @property
    def train_sample_count(self) -> int:
        """the number of entries in the table that was used to train Aito during evaluation

        :rtype: int
        """
        return self.__getitem__('trainSamples')



import json
import sys
from typing import Dict
from typing import List

from aito.schema import AitoTableSchema
from aito.utils.data_frame_handler import DataFrameHandler
from .sub_command import SubCommand
from ..parser import InputArgType, ParseError, create_sql_connecting_from_parsed_args


class InferFromFormatSubCommand(SubCommand):
    def build_parser(self, parser):
        parser.add_argument(
            'input', default='-', type=InputArgType(), nargs='?',
            help="path to the input file (when no input file is given or when input is -, read from the standard input)"
        )

    @staticmethod
    def parsed_args_to_data_frame_handler_read_args(parsed_args: Dict) -> Dict:
        in_format = parsed_args['input-format']

        read_args = {
            'read_input': parsed_args['input'],
            'in_format': in_format,
            'read_options': {}
        }

        if in_format == 'csv':
            read_args['read_options']['delimiter'] = parsed_args['delimiter']
            read_args['read_options']['decimal'] = parsed_args['decimal']
        elif in_format == 'excel':
            if parsed_args['input'] == sys.stdin:
                raise ParseError('input must be a file path for excel files')
            if parsed_args['one_sheet']:
                read_args['read_options']['sheet_name'] = parsed_args['one_sheet']
        return read_args

    def parse_and_execute(self, parsed_args: Dict):
        parsed_read_args = self.parsed_args_to_data_frame_handler_read_args(parsed_args)
        df = DataFrameHandler().read_file_to_df(**parsed_read_args)
        inferred_schema = AitoTableSchema.infer_from_pandas_data_frame(df)
        json.dump(inferred_schema.to_json_serializable(), sys.stdout, indent=4, sort_keys=True)
        return 0


class InferFromCSVSubCommand(InferFromFormatSubCommand):
    def __init__(self):
        super().__init__('csv', 'infer a table schema from CSV data')

    def build_parser(self, parser):
        super().build_parser(parser)
        parser.add_csv_format_default_arguments()


class InferFromExcelSubCommand(InferFromFormatSubCommand):
    def __init__(self):
        super().__init__('excel', 'infer a table schema from EXCEL data')

    def build_parser(self, parser):
        super().build_parser(parser)
        parser.add_excel_format_default_arguments()


class InferFromSQLSubCommand(SubCommand):
    def __init__(self):
        super().__init__('from-sql', 'infer a table schema from the result of a SQL query')

    def build_parser(self, parser):
        parser.add_sql_default_credentials_arguments()
        parser.add_argument('query', type=str, help='query to get the data from your database')

    def parse_and_execute(self, parsed_args: Dict):
        connection = create_sql_connecting_from_parsed_args(parsed_args)
        result_df = connection.execute_query_and_save_result(parsed_args['query'])
        inferred_schema = AitoTableSchema.infer_from_pandas_data_frame(result_df)
        json.dump(inferred_schema.to_json_serializable(), sys.stdout, indent=4, sort_keys=True)
        return 0


class InferTableSchemaSubCommand(SubCommand):
    _default_sub_commands = [
        InferFromCSVSubCommand(),
        InferFromExcelSubCommand(),
        InferFromFormatSubCommand('json', 'infer a table schema from JSON data'),
        InferFromFormatSubCommand('ndjson', 'infer a table schema from NDJSON data'),
        InferFromSQLSubCommand()
    ]

    def __init__(self, sub_commands: List[SubCommand] = None):
        super().__init__('infer-table-schema', 'infer an Aito table schema from a file')
        if not sub_commands:
            sub_commands = self._default_sub_commands
        self._sub_commands_map = {cmd.name: cmd for cmd in sub_commands}

    def build_parser(self, parser):
        parser.epilog = '''To see help for a specific format:
  aito infer-table-schema <input-format> - h

When no input or when input is -, read standard input.
You must use input file instead of standard input for excel file
'''
        sub_commands_subparsers = parser.add_subparsers(
            title='input-format',
            dest='input-format',
            metavar='<input-format>'
        )
        sub_commands_subparsers.required = True

        for sub_cmd in self._sub_commands_map.values():
            sub_cmd_parser = sub_commands_subparsers.add_parser(sub_cmd.name, help=sub_cmd.help_message)
            sub_cmd.build_parser(sub_cmd_parser)

    def parse_and_execute(self, parsed_args: Dict):
        self._sub_commands_map[parsed_args['input-format']].parse_and_execute(parsed_args)
        return 0



"""Aito `Job API <https://aito.ai/docs/api/#post-api-v1-jobs-query>`__ Request Class"""

import re
from abc import abstractmethod
from typing import Optional, Union, Dict, List, Type

from aito.client import responses as aito_resp
from .aito_request import AitoRequest, _PatternEndpoint, _GetRequest, _PostRequest
from .query_api_request import QueryAPIRequest


class _JobAPIRequest(AitoRequest):
    endpoint_prefix = f'{AitoRequest._api_version_endpoint_prefix}/jobs'

    @classmethod
    @abstractmethod
    def _check_method(cls, method: str) -> bool:
        return method in cls._request_methods

    @classmethod
    @abstractmethod
    def _check_endpoint(cls, endpoint: str) -> bool:
        return endpoint.startswith(cls.endpoint_prefix)

    @classmethod
    @abstractmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        for sub_cls in cls.__subclasses__():
            if sub_cls._check_method(method) and sub_cls._check_endpoint(endpoint):
                return sub_cls.make_request(method=method, endpoint=endpoint, query=query)
        raise ValueError(f"invalid {cls.__name__} with '{method}({endpoint})'")


class CreateJobRequest(_PostRequest, _PatternEndpoint, _JobAPIRequest):
    """Request to `create a job <https://aito.ai/docs/api/#post-api-v1-jobs-query>`__ for an Aito API endpoint
    """
    response_cls = aito_resp.CreateJobResponse

    def __init__(self, endpoint: str, query: Dict):
        """

        :param endpoint: the job endpoint
        :type endpoint: str
        :param query: the query
        :type query: Dict
        """
        super().__init__(method=self.method, endpoint=endpoint, query=query)

    @property
    def path(self):
        return self.endpoint.replace(f'{self.endpoint_prefix}/', '')

    @classmethod
    def _endpoint_pattern(cls):
        return re.compile(f"^{cls.endpoint_prefix}/({'|'.join(QueryAPIRequest._query_api_paths)})$")

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        return cls(endpoint=endpoint, query=query)

    @classmethod
    def from_query_api_request(cls, request_obj: QueryAPIRequest) -> 'CreateJobRequest':
        """Create a job from a QueryAPI request

        :param request_obj: a :class:`.QueryAPIRequest` instance
        :type request_obj: QueryAPIRequest
        :return: the corresponding create job request
        :rtype: CreateJobRequest
        """
        endpoint = f'{cls.endpoint_prefix}/{request_obj.path}'
        return cls(endpoint=endpoint, query=request_obj.query)

    @property
    def result_response_cls(self) -> Type[aito_resp.BaseResponse]:
        """returns the response class of the job request result"""
        for sub_cls in QueryAPIRequest.__subclasses__():
            if self.path == sub_cls.path:
                return sub_cls.response_cls


class GetJobStatusRequest(_GetRequest, _PatternEndpoint, _JobAPIRequest):
    """Request to `create a job <https://aito.ai/docs/api/#post-api-v1-jobs-query>`__ for an Aito API endpoint
    """
    response_cls = aito_resp.GetJobStatusResponse

    def __init__(self, job_id: str):
        """

        :param job_id: the id of the job session
        :type job_id: str
        """
        endpoint = f'{self.endpoint_prefix}/{job_id}'
        super().__init__(method=self.method, endpoint=endpoint)

    @classmethod
    def _endpoint_pattern(cls):
        return re.compile(f'^{cls.endpoint_prefix}/(.+)$')

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        groups = cls._endpoint_to_captured_groups(endpoint=endpoint)
        return cls(job_id=groups(1))


class GetJobResultRequest(_GetRequest, _PatternEndpoint, _JobAPIRequest):
    """Request to `create a job <https://aito.ai/docs/api/#post-api-v1-jobs-query>`__ for an Aito API endpoint
    """
    response_cls = aito_resp.BaseResponse

    def __init__(self, job_id: str):
        """

        :param job_id: the id of the job session
        :type job_id: str
        """
        endpoint = f'{self.endpoint_prefix}/{job_id}/result'
        super().__init__(method=self.method, endpoint=endpoint)

    @classmethod
    def _endpoint_pattern(cls):
        return re.compile(f'^{cls.endpoint_prefix}/(.+)/result$')

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        groups = cls._endpoint_to_captured_groups(endpoint=endpoint)
        return cls(job_id=groups(1))


"""A utility to read, write, and convert a Pandas DataFrame in accordance to a Aito Table Schema
"""

import logging
from typing import List, Dict, Callable

import pandas as pd

from aito.utils._typing import *
from aito.schema import AitoTableSchema, DataSeriesProperties

LOG = logging.getLogger('DataFrameHandler')


class DataFrameHandler:
    """A handler that supports read, write, and convert a Pandas DataFrame in accordance to a Aito Table Schema
    """
    allowed_format = ['csv', 'json', 'excel', 'ndjson']

    def __init__(self):
        self.default_options = {
            'csv': {'index_col':False, 'error_bad_lines':False, 'warn_bad_lines':True, 'engine':'python'},
            'excel': {},
            'json': {'orient': 'records'},
            'ndjson': {'orient': 'records', 'lines': True}
        }
        self.default_apply_functions = [self._datetime_to_string]

    def _validate_in_out_format(self, in_format: str, out_format: str):
        """Validate the file parameters of the converter
        """
        if in_format not in self.allowed_format:
            raise ValueError(
                f"Expect the input format to be one of {'|'.join(self.allowed_format)} instead of {in_format}"
            )

        if out_format not in self.allowed_format:
            raise ValueError(
                f"Expect the output format to be one of {'|'.join(self.allowed_format)} instead of {out_format}"
            )

    @staticmethod
    def _datetime_to_string(df: pd.DataFrame) -> pd.DataFrame:
        """Convert pandas datetime type to string
        """
        for col in df:
            if df[col].dtypes == 'datetime64[ns]':
                df[col] = df[col].astype(str)
        return df

    @staticmethod
    def _apply_functions_on_df(df: pd.DataFrame, functions: List[Callable]) -> pd.DataFrame:
        """Applying partial functions to a dataframe

        :param df: input pandas DataFrame
        :type df: pd.DataFrame
        :param functions: list of partial functions that will be applied to the loaded pd.DataFrame
        :type functions: List[Callable]
        :return: output DataFrame
        :rtype: pd.DataFrame
        """
        for f in functions:
            df = f(df)
        return df

    @staticmethod
    def convert_df_using_aito_table_schema(
            df: pd.DataFrame, table_schema: Union[AitoTableSchema, Dict]
    ) -> pd.DataFrame:
        """convert a pandas DataFrame to match a given Aito table schema

        :param df: input pandas DataFrame
        :type df: pd.DataFrame
        :param table_schema: input table schema
        :type table_schema: an AitoTableSchema object or a Dict, optional
        :raises ValueError: input table schema is invalid
        :raises e: failed to convert
        :return: converted DataFrame
        :rtype: pd.DataFrame
        """
        if not isinstance(table_schema, AitoTableSchema):
            if not isinstance(table_schema, dict):
                raise ValueError("the input table schema must be either an AitoTableSchema object or a dict")
            table_schema = AitoTableSchema.from_deserialized_object(table_schema)

        df_columns = set(df.columns.values)
        table_schema_columns = set(table_schema.columns)

        for col_name in (df_columns - table_schema_columns):
            LOG.warning(f"column `{col_name}` found in the input data but not found in the input schema")
        for col_name in (table_schema_columns - df_columns):
            LOG.warning(f"column `{col_name}` found in the input schema but not found in the input data")

        conversion_map = {}
        for col_name in table_schema_columns.intersection(df_columns):
            col_schema = table_schema[col_name]
            col_df_nullable = df[col_name].isna().any()
            if col_df_nullable and not col_schema.nullable:
                raise ValueError(f"column `{col_name}` is nullable but stated non-nullable in the input schema")
            conversion_map[col_name] = col_schema.to_conversion()

        LOG.debug(f"casting dataframe columns: {conversion_map}")
        converted_df = df
        for col_name in conversion_map:
            conversion = conversion_map[col_name]
            converted_df[col_name] = converted_df[col_name].apply(conversion)

        LOG.debug(f"converted the dataframe according to the schema")
        return converted_df

    def read_file_to_df(self, read_input: FilePathOrBuffer, in_format: str, read_options: Dict = None) -> pd.DataFrame:
        """Read input to a Pandas DataFrame

        :param read_input: read input
        :type read_input: any valid string path, pathlike object, or file-like object (objects with a read() method)
        :param in_format: input format
        :type in_format: str
        :param read_options: dictionary contains arguments for pandas read function, defaults to None
        :type read_options: Dict, optional
        :return: read DataFrame
        :rtype: pd.DataFrame
        """
        LOG.debug(f'reading data from {read_input} to df...')
        read_functions = {'csv': pd.read_csv, 'excel': pd.read_excel, 'json': pd.read_json, 'ndjson': pd.read_json}

        if not read_options:
            options = self.default_options[in_format]
        else:
            options = read_options
            options.update(self.default_options[in_format])
        df = read_functions[in_format](read_input, **options)
        return df

    def df_to_format(
            self,
            df: pd.DataFrame,
            out_format: str,
            write_output: FilePathOrBuffer,
            convert_options: Dict = None
    ):
        """Write a Pandas DataFrame

        :param df: input DataFrame
        :type df: pd.DataFrame
        :param out_format: output format
        :type out_format: str
        :param write_output: write output
        :type write_output: any valid string path, pathlike object, or file-like object (objects with a read() method)
        :param convert_options: dictionary contains arguments for pandas write function, defaults to None
        :type convert_options: Dict, optional
        """
        convert_functions = {'csv': df.to_csv, 'excel': df.to_excel, 'json': df.to_json, 'ndjson': df.to_json}
        if not convert_options:
            options = self.default_options[out_format]
        else:
            options = convert_options
            options.update(self.default_options[out_format])

        LOG.debug(f'converting to {out_format} and writing to {write_output} with options {options}...')

        convert_functions[out_format](write_output, **options)

    def convert_file(
            self,
            read_input: FilePathOrBuffer,
            write_output: FilePathOrBuffer,
            in_format: str,
            out_format: str,
            read_options: Dict = None,
            convert_options: Dict = None,
            apply_functions: List[Callable[..., pd.DataFrame]] = None,
            use_table_schema: Union[AitoTableSchema, Dict] = None
    ) -> pd.DataFrame:
        """Converting input file to expected format, generate or use Aito table schema if specified

        :param read_input: read input
        :type read_input: any valid string path, pathlike object, or file-like object (objects with a read() method)
        :param write_output: write output
        :type write_output: any valid string path, pathlike object, or file-like object (objects with a read() method)
        :param in_format: input format
        :type in_format: str
        :param out_format: output format
        :type out_format: str
        :param read_options: dictionary contains arguments for pandas read function, defaults to None
        :type read_options: Dict, optional
        :param convert_options: dictionary contains arguments for pandas write function, defaults to None
        :type convert_options: Dict, optional
        :param apply_functions: list of partial functions that will be applied to the loaded pd.DataFrame, defaults to None
        :type apply_functions: List[Callable[..., pd.DataFrame]], optional
        :param use_table_schema: use an aito schema to dictates data types and convert the data, defaults to None
        :type use_table_schema: an AitoTableSchema object or a Dict, optional
        :return: converted DataFrame
        :rtype: pd.DataFrame
        """
        self._validate_in_out_format(in_format, out_format)

        df = self.read_file_to_df(read_input, in_format, read_options)

        if apply_functions:
            apply_functions = self.default_apply_functions + apply_functions
        else:
            apply_functions = self.default_apply_functions
        df = self._apply_functions_on_df(df, apply_functions)

        # if no schema is predefined, infer the schema and use the inferred schema
        used_table_schema = use_table_schema
        if used_table_schema is None:
            used_table_schema = AitoTableSchema.infer_from_pandas_data_frame(df)

        df = self.convert_df_using_aito_table_schema(df, used_table_schema)

        if out_format != in_format or convert_options or use_table_schema:
            self.df_to_format(df, out_format, write_output, convert_options)
        return df



"""Aito `Schema API <https://aito.ai/docs/api/#database-api>`__ Request Class"""

import re
from abc import ABC, abstractmethod
from typing import Optional, Union, Dict, List

from aito.client import responses as aito_responses
from .aito_request import AitoRequest, _PatternEndpoint, _GetRequest, _PutRequest, _DeleteRequest
from aito.schema import AitoDatabaseSchema, AitoTableSchema


class _SchemaAPIRequest(AitoRequest, ABC):
    """Request to manipulate the schema"""
    endpoint_prefix = f'{AitoRequest._api_version_endpoint_prefix}/schema'

    @classmethod
    @abstractmethod
    def _check_method(cls, method: str) -> bool:
        return method in cls._request_methods

    @classmethod
    @abstractmethod
    def _check_endpoint(cls, endpoint: str) -> bool:
        return endpoint.startswith(cls.endpoint_prefix)

    @classmethod
    @abstractmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        for sub_cls in cls.__subclasses__():
            if sub_cls._check_method(method) and sub_cls._check_endpoint(endpoint):
                return sub_cls.make_request(method=method, endpoint=endpoint, query=query)
        raise ValueError(f"invalid {cls.__name__} with '{method}({endpoint})'")


class _DatabaseSchemaRequest:
    """Request to manipulate the database schema"""
    endpoint = _SchemaAPIRequest.endpoint_prefix

    @classmethod
    def _check_endpoint(cls, endpoint: str):
        return endpoint == _SchemaAPIRequest.endpoint_prefix


class _TableSchemaRequest(_PatternEndpoint):
    """Request to manipulate a table schema"""
    @classmethod
    def _endpoint_pattern(cls):
        return re.compile(f'^{_SchemaAPIRequest.endpoint_prefix}/([^/".$\r\n\s]+)$')


class _ColumnSchemaRequest(_PatternEndpoint):
    """Request to manipulate a column schema"""
    @classmethod
    def _endpoint_pattern(cls):
        return re.compile(f'^{_SchemaAPIRequest.endpoint_prefix}/([^/".$\r\n\s]+)/([^/".$\r\n\s]+)$')


class GetDatabaseSchemaRequest(_GetRequest, _DatabaseSchemaRequest, _SchemaAPIRequest):
    """Request to `Get the schema of the database <https://aito.ai/docs/api/#get-api-v1-schema>`__"""
    response_cls = aito_responses.DatabaseSchemaResponse

    def __init__(self):
        super().__init__(method=self.method, endpoint=self.endpoint)

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        return cls()


class GetTableSchemaRequest(_GetRequest, _TableSchemaRequest, _SchemaAPIRequest):
    """Request to `Get the schema of a table <https://aito.ai/docs/api/#get-api-v1-schema-table>`__"""
    response_cls = aito_responses.TableSchemaResponse

    def __init__(self, table_name: str):
        """

        :param table_name: the name of the table
        :type table_name: str
        """
        endpoint = f'{self.endpoint_prefix}/{table_name}'
        super().__init__(method=self.method, endpoint=endpoint)

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        groups = cls._endpoint_to_captured_groups(endpoint=endpoint)
        return cls(table_name=groups(1))


class GetColumnSchemaRequest(_GetRequest, _ColumnSchemaRequest, _SchemaAPIRequest):
    """Request to `Get the schema of a column <https://aito.ai/docs/api/#get-api-v1-schema-column>`__"""
    response_cls = aito_responses.ColumnSchemaResponse

    def __init__(self, table_name: str, column_name: str):
        """

        :param table_name: the name of the table containing the column
        :type table_name: str
        :param column_name: the name of the column
        :type column_name: str
        """
        endpoint = f'{self.endpoint_prefix}/{table_name}/{column_name}'
        super().__init__(method=self.method, endpoint=endpoint)

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        groups = cls._endpoint_to_captured_groups(endpoint=endpoint)
        return cls(table_name=groups(1), column_name=groups(2))


class CreateDatabaseSchemaRequest(_PutRequest, _DatabaseSchemaRequest, _SchemaAPIRequest):
    """Request to `Create the schema of the database <https://aito.ai/docs/api/#put-api-v1-schema>`__"""
    endpoint = _SchemaAPIRequest.endpoint_prefix
    response_cls = aito_responses.DatabaseSchemaResponse

    def __init__(self, schema: Union[AitoDatabaseSchema, Dict]):
        """

        :param schema: Aito database schema
        :type schema: Union[AitoDatabaseSchema, Dict]
        """
        query = schema.to_json_serializable() if isinstance(schema, AitoDatabaseSchema) else schema
        super().__init__(method=self.method, endpoint=self.endpoint, query=query)

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        return cls(schema=query)


class CreateTableSchemaRequest(_PutRequest, _TableSchemaRequest, _SchemaAPIRequest):
    """Request to `Create a table <https://aito.ai/docs/api/#put-api-v1-schema-table>`__"""
    response_cls = aito_responses.schema_api_response.TableSchemaResponse

    def __init__(self, table_name: str, schema: Union[AitoTableSchema, Dict]):
        """

        :param table_name: the name of the table
        :type table_name: str
        :param schema: the schema of the table
        :type schema: Union[AitoDatabaseSchema, Dict]
        """
        endpoint = f'{self.endpoint_prefix}/{table_name}'
        query = schema.to_json_serializable() if isinstance(schema, AitoTableSchema) else schema
        super().__init__(method=self.method, endpoint=endpoint, query=query)

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        groups = cls._endpoint_to_captured_groups(endpoint=endpoint)
        return cls(table_name=groups(1), schema=query)


class CreateColumnSchemaRequest(_PutRequest, _ColumnSchemaRequest, _SchemaAPIRequest):
    """Request to `Add or replace a column <https://aito.ai/docs/api/#put-api-v1-schema-column>`__"""
    response_cls = aito_responses.ColumnSchemaResponse

    def __init__(self, table_name: str, column_name: str, schema: Dict):
        """

        :param table_name: the name of the table containing the column
        :type table_name: str
        :param column_name: the name of the column
        :type column_name: str
        :param schema: the schema of the column
        :type schema: Union[AitoColumnTypeSchema, Dict]
        """
        endpoint = f'{self.endpoint_prefix}/{table_name}/{column_name}'
        super().__init__(method=self.method, endpoint=endpoint, query=schema)

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        groups = cls._endpoint_to_captured_groups(endpoint=endpoint)
        return cls(table_name=groups(1), column_name=groups(2), schema=query)


class DeleteDatabaseSchemaRequest(_DeleteRequest, _DatabaseSchemaRequest, _SchemaAPIRequest):
    """Request to `Delete the schema of the database <https://aito.ai/docs/api/#delete-api-v1-schema>`__"""
    endpoint = _SchemaAPIRequest.endpoint_prefix
    response_cls = aito_responses.BaseResponse

    def __init__(self):
        super().__init__(method=self.method, endpoint=self.endpoint)

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        return cls()


class DeleteTableSchemaRequest(_DeleteRequest, _TableSchemaRequest, _SchemaAPIRequest):
    """Request to `Delete a table <https://aito.ai/docs/api/#delete-api-v1-schema-table>`__"""
    response_cls = aito_responses.BaseResponse

    def __init__(self, table_name: str):
        """

        :param table_name: the name of the table
        :type table_name: str
        """
        endpoint = f'{self.endpoint_prefix}/{table_name}'
        super().__init__(method=self.method, endpoint=endpoint)

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        groups = cls._endpoint_to_captured_groups(endpoint=endpoint)
        return cls(table_name=groups(1))


class DeleteColumnSchemaRequest(_DeleteRequest, _ColumnSchemaRequest, _SchemaAPIRequest):
    """Request to `Delete a column <https://aito.ai/docs/api/#delete-api-v1-schema-column>`__"""
    response_cls = aito_responses.BaseResponse

    def __init__(self, table_name: str, column_name: str):
        """

        :param table_name: the name of the table containing the column
        :type table_name: str
        :param column_name: the name of the column
        :type column_name: str
        """
        endpoint = f'{self.endpoint_prefix}/{table_name}/{column_name}'
        super().__init__(method=self.method, endpoint=endpoint)

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        groups = cls._endpoint_to_captured_groups(endpoint=endpoint)
        return cls(table_name=groups(1), column_name=groups(2))


"""A versatile client that makes requests to an Aito Database Instance

"""

import asyncio
import logging
import warnings
from typing import Dict, List, Union, Tuple, Optional

import requests as requestslib
from aiohttp import ClientSession, ClientResponseError

from aito.exceptions import BaseError
from .requests import AitoRequest, BaseRequest, GetVersionRequest
from .responses import BaseResponse

LOG = logging.getLogger('AitoClient')


class Error(BaseError):
    """An error occurred when using the client

    """
    def __init__(self, message: str):
        super().__init__(message, LOG)


class RequestError(Error):
    """An error occurred when sending a request to the Aito instance

    """
    def __init__(self, request_obj: AitoRequest, error: Exception):
        """

        :param request_obj: the request object
        :type request_obj: AitoRequest
        :param error: the error
        :type error: Exception
        """
        self.request_obj = request_obj
        self.error = error
        if isinstance(error, requestslib.HTTPError):
            resp = error.response.json()
            error_msg = resp['message'] if 'message' in resp else resp
        elif isinstance(error, ClientResponseError):
            error_msg = error.message
        else:
            error_msg = str(error)
        super().__init__(f'failed to {request_obj}: {error_msg}')


class AitoClient:
    """A versatile client that connects to the Aito Database Instance

    """
    def __init__(
            self,
            instance_url: str,
            api_key: str,
            check_credentials: bool = True,
            raise_for_status: bool = True
    ):
        """

        :param instance_url: Aito instance url
        :type instance_url: str
        :param api_key: Aito instance API key
        :type api_key: str
        :param check_credentials: check the given credentials by requesting the Aito instance version, defaults to True
        :type check_credentials: bool
        :param raise_for_status: automatically raise RequestError for each failed response, defaults to True
        :type raise_for_status: bool
        :raises BaseError: an error occurred during the creation of AitoClient

        >>> aito_client = AitoClient(your_instance_url, your_api_key) # doctest: +SKIP
        >>> # Change the API key to READ-WRITE or READ-ONLY
        >>> aito_client.api_key = new_api_key # doctest: +SKIP
        """
        self.instance_url = instance_url.strip("/")
        self.api_key = api_key
        self.raise_for_status = raise_for_status
        self.instance_version = None
        if check_credentials:
            try:
                version_resp = self.request(request_obj=GetVersionRequest(), raise_for_status=True)
                self.instance_version = version_resp.version
            except Exception:
                raise Error('failed to instantiate Aito Client, please check your credentials')

    @property
    def headers(self):
        """ the headers that will be used to send a request to the Aito instance

        :rtype: Dict
        """
        return {'Content-Type': 'application/json', 'x-api-key': self.api_key}

    def request(
            self, *,
            method: Optional[str] = None,
            endpoint: Optional[str] = None,
            query: Optional[Union[Dict, List]] = None,
            request_obj: Optional[AitoRequest] = None,
            raise_for_status: Optional[bool] = None
    ) -> Union[BaseResponse, RequestError]:
        # noinspection LongLine
        """make a request to an Aito API endpoint
        The client returns a JSON response if the request succeed and a :class:`.RequestError` if the request fails

        :param method: method for the new :class:`.AitoRequest` object
        :type method: str
        :param endpoint: endpoint for the new :class:`.AitoRequest` object
        :type endpoint: str
        :param query: an Aito query if applicable, defaults to None
        :type query: Optional[Union[Dict, List]]
        :param request_obj: use an :class:`.AitoRequest` object
        :type request_obj: AitoRequest
        :param raise_for_status: raise :class:`.RequestError` if the request fails instead of returning the error
            If set to None, value from Client will be used. Defaults to True
        :type raise_for_status: Optional[bool]
        :raises RequestError: an error occurred during the execution of the request and raise_for_status
        :return: a response object or :class:`.RequestError` if an error occurred and not raise_for_status
        :rtype: Union[BaseResponse, RequestError]

        Simple request to get the schema of a table:

        >>> res = client.request(method="GET", endpoint="/api/v1/schema/impressions")
        >>> print(res.to_json_string(indent=2, sort_keys=True))
        {
          "columns": {
            "product": {
              "link": "products.id",
              "nullable": false,
              "type": "String"
            },
            "purchase": {
              "nullable": false,
              "type": "Boolean"
            },
            "session": {
              "link": "sessions.id",
              "nullable": false,
              "type": "String"
            }
          },
          "type": "table"
        }

         Sends a `PREDICT <https://aito.ai/docs/api/#post-api-v1-predict>`__ query:

         >>> from aito.client import PredictRequest
         >>> res = client.request(request_obj=PredictRequest(
         ...    query={
         ...        "from": "impressions",
         ...        "where": { "session": "veronica" },
         ...        "predict": "product.name"
         ...    }
         ... )) # doctest: +NORMALIZE_WHITESPACE
         >>> print(res.top_prediction) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
         {"$p": ..., "field": ..., "feature": ...}

         Returns an error when make a request to an incorrect path:

         >>> client.request(method="GET", endpoint="api/v1/incorrect-path") # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
         Traceback (most recent call last):
            ...
         ValueError: invalid endpoint 'api/v1/incorrect-path' for BaseRequest
         """
        if request_obj is None:
            if method is not None and endpoint is not None:
                request_obj = AitoRequest.make_request(method, endpoint, query)
            else:
                raise TypeError("'request() requires either 'request_obj' or 'method' and 'endpoint'")
        try:
            resp = requestslib.request(
                method=request_obj.method,
                url=self.instance_url + request_obj.endpoint,
                headers=self.headers,
                json=request_obj.query
            )
            resp.raise_for_status()
            json_resp = resp.json()
        except Exception as e:
            req_err = RequestError(request_obj, e)
            _raise = raise_for_status if raise_for_status is not None else self.raise_for_status
            if _raise:
                raise req_err
            else:
                return req_err
        return request_obj.response_cls(json_resp)

    async def async_request(
            self, session: ClientSession, *,
            method: Optional[str] = None,
            endpoint: Optional[str] = None,
            query: Optional[Union[Dict, List]] = None,
            request_obj: Optional[AitoRequest] = None,
            raise_for_status: Optional[bool] = None
    ) -> Union[BaseResponse, RequestError]:
        """execute a request asynchronously using aiohttp ClientSession

        :param session: aiohttp ClientSession for making request
        :type session: ClientSession
        :param method: method for the new :class:`.AitoRequest` object
        :type method: str
        :param endpoint: endpoint for the new :class:`.AitoRequest` object
        :type endpoint: str
        :param query: an Aito query if applicable, defaults to None
        :type query: Optional[Union[Dict, List]]
        :param request_obj: use an :class:`.AitoRequest` object
        :type request_obj: AitoRequest
        :param raise_for_status: raise :class:`.RequestError` if the request fails instead of returning the error
            If set to None, value from Client will be used. Defaults to True
        :type raise_for_status: Optional[bool]
        :raises RequestError: an error occurred during the execution of the request and raise_for_status
        """
        LOG.debug(f'async {request_obj}')
        if request_obj is None:
            if method is not None and endpoint is not None:
                request_obj = AitoRequest.make_request(method, endpoint, query)
            else:
                raise TypeError("request() requires either 'request_obj' or 'method' and 'endpoint'")
        try:
            async with session.request(
                    method=request_obj.method,
                    url=self.instance_url + request_obj.endpoint,
                    json=request_obj.query,
                    headers=self.headers,
                    raise_for_status=True
            ) as resp:
                return request_obj.response_cls(await resp.json())
        except Exception as e:
            req_err = RequestError(request_obj, e)
            _raise = raise_for_status if raise_for_status is not None else self.raise_for_status
            if _raise:
                raise req_err
            else:
                return req_err

    async def bounded_async_request(
            self, semaphore: asyncio.Semaphore, *args, **kwargs
    ) -> Union[BaseResponse, RequestError]:
        """bounded concurrent requests with asyncio semaphore

        :param semaphore: asyncio Semaphore
        :type semaphore: asyncio.Semaphore
        :param args: :func:`.async_request` positional arguments
        :param kwargs: :func:`.async_request` keyword arguments
        :return: tuple of request status code and request json content
        :rtype: Tuple[int, Union[Dict, List]]
        """
        async with semaphore:
            return await self.async_request(*args, **kwargs)

    def async_requests(
            self,
            methods: List[str],
            endpoints: List[str],
            queries: List[Union[List, Dict]],
            batch_size: int = 10
    ) -> List[BaseResponse]:
        """
        .. deprecated:: 0.4.0

        Use :func:`batch_requests` instead
        """
        warnings.warn(
            'The AitoClient.async_requests function is deprecated and will be removed '
            'in a future version. Use AitoClient.batch_requests instead',
            category=FutureWarning
        )
        requests = [BaseRequest(method, endpoints[idx], queries[idx]) for idx, method in enumerate(methods)]
        return self.batch_requests(requests=requests, max_concurrent_requests=batch_size)

    def batch_requests(
            self,
            requests: List[AitoRequest],
            max_concurrent_requests: int = 10
    ) -> List[Union[BaseResponse, RequestError]]:
        """execute a batch of requests asynchronously

        This method is useful when sending a batch of requests, for example, when sending a batch of predict requests.

        :param requests: list of request objects
        :type requests: List[AitoRequest]
        :param max_concurrent_requests: the number of queries to be sent per batch
        :type max_concurrent_requests: int
        :return: list of request response or exception if a request did not succeed
        :rtype: List[Union[BaseResponse, RequestError]]

        Find products that multiple users would most likely buy

        >>> from aito.client import MatchRequest
        >>> users = ['veronica', 'larry', 'alice']
        >>> responses = client.batch_requests([
        ...     MatchRequest(
        ...         query = {
        ...             'from': 'impressions',
        ...             'where': { 'session.user': usr },
        ...             'match': 'product'
        ...         }
        ...     )
        ...     for usr in users
        ... ])
        >>> # Print top product for each customer
        >>> for idx, usr in enumerate(users):
        ...     print(f"{usr}: {responses[idx].top_match}") # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
        veronica: {"$p": ..., "category": ..., "id": ..., "name": ..., "price": ..., "tags": ...}
        larry: {"$p": ..., "category": ..., "id": ..., "name": ..., "price": ..., "tags": ...}
        alice: {"$p": ..., "category": ..., "id": ..., "name": ..., "price": ..., "tags": ...}

        """
        async def run():
            async with ClientSession() as session:
                tasks = [
                    self.bounded_async_request(semaphore, session=session, request_obj=req, raise_for_status=False)
                    for req in requests
                ]
                return await asyncio.gather(*tasks)

        semaphore = asyncio.Semaphore(max_concurrent_requests)
        loop = asyncio.get_event_loop()
        responses = loop.run_until_complete(run())
        return responses



"""Aito `Job API <https://aito.ai/docs/api/#post-api-v1-jobs-query>`__ Response Class"""

from typing import Optional

from .aito_response import BaseResponse


class CreateJobResponse(BaseResponse):
    """Response of the `Create job <https://aito.ai/docs/api/#post-api-v1-jobs-query>`__
    containing the job information
    """
    @property
    def id(self) -> str:
        """the id of a job"""
        return self.__getitem__('id')

    @property
    def started_at(self) -> str:
        """when the job was started"""
        return self.__getitem__('startedAt')


class GetJobStatusResponse(BaseResponse):
    """Response of the `Get job status <https://aito.ai/docs/api/#get-api-v1-jobs-uuid>`__"""

    @property
    def id(self) -> str:
        """the id of the job session"""
        return self.__getitem__('id')

    @property
    def started_at(self) -> str:
        """when the job was started"""
        return self.__getitem__('startedAt')

    @property
    def finished(self) -> bool:
        """if the job has finished"""
        return self.__contains__('finishedAt')

    @property
    def expires_at(self) -> Optional[str]:
        """when the job result will not be available, returns None if job hasn't finished"""
        return self.__getitem__('startedAt') if self.finished else None


"""Different APIs that takes an :class:`Aito Client object <aito.client.AitoClient>` as the first argument

"""

import logging
import tempfile
import time
from os import PathLike, unlink
from pathlib import Path
from typing import Dict, List, BinaryIO, Union, Tuple, Iterable, Optional

import ndjson
import requests as requestslib

import aito.client.requests as aito_requests
import aito.client.responses as aito_responses
from aito.client import AitoClient, RequestError
from aito.schema import AitoDatabaseSchema, AitoTableSchema, AitoColumnTypeSchema
from aito.utils._file_utils import gzip_file, check_file_is_gzipped
from aito.utils.data_frame_handler import DataFrameHandler

LOG = logging.getLogger('AitoAPI')


def get_version(client: AitoClient) -> str:
    """get the aito instance version

    :param client: the AitoClient instance
    :type client: AitoClient
    :return: version information in json format
    :rtype: Dict
    """
    resp = client.request(request_obj=aito_requests.GetVersionRequest())
    return resp.version


def create_database(client: AitoClient, schema: Union[AitoDatabaseSchema, Dict]):
    """`create a database <https://aito.ai/docs/api/#put-api-v1-schema>`__ using the specified database schema

    .. note::

        requires the client to be setup with the READ-WRITE API key

    :param client: the AitoClient instance
    :type client: AitoClient
    :param schema: the schema of the database
    :type schema: Dict
    """
    client.request(request_obj=aito_requests.CreateDatabaseSchemaRequest(schema=schema))
    LOG.info('database schema created')


def get_database_schema(client: AitoClient) -> AitoDatabaseSchema:
    """`get the schema of the database <https://aito.ai/docs/api/#get-api-v1-schema>`__

    :param client: the AitoClient instance
    :type client: AitoClient
    :return: Aito database schema
    :rtype: Dict
    """
    res = client.request(request_obj=aito_requests.GetDatabaseSchemaRequest())
    return res.schema


def delete_database(client: AitoClient):
    """`delete the whole database <https://aito.ai/docs/api/#delete-api-v1-schema>`__

    .. note::

        requires the client to be setup with the READ-WRITE API key

    :param client: the AitoClient instance
    :type client: AitoClient
    """
    client.request(request_obj=aito_requests.DeleteDatabaseSchemaRequest())
    LOG.info('database deleted')


def create_table(client: AitoClient, table_name: str, schema: Union[AitoTableSchema, Dict]):
    """`create a table <https://aito.ai/docs/api/#put-api-v1-schema-table>`__
    with the specified table name and schema

    update the table if the table already exists and does not contain any data

    .. note::

        requires the client to be setup with the READ-WRITE API key

    :param client: the AitoClient instance
    :type client: AitoClient
    :param table_name: the name of the table
    :type table_name: str
    :param schema: Aito table schema
    :type schema: an AitoTableSchema object or a Dict, optional
    """
    client.request(request_obj=aito_requests.CreateTableSchemaRequest(table_name=table_name, schema=schema))
    LOG.info(f'table `{table_name}` created')


def get_table_schema(client: AitoClient, table_name: str) -> AitoTableSchema:
    """`get the schema of the specified table <https://aito.ai/docs/api/#get-api-v1-schema-table>`__

    :param client: the AitoClient instance
    :type client: AitoClient
    :param table_name: the name of the table
    :type table_name: str
    :return: the table schema
    :rtype: AitoTableSchema
    """
    resp = client.request(request_obj=aito_requests.GetTableSchemaRequest(table_name=table_name))
    return resp.schema


def delete_table(client: AitoClient, table_name: str):
    """`delete the specified table <https://aito.ai/docs/api/#delete-api-v1-schema>`__

    .. note::

        requires the client to be setup with the READ-WRITE API key

    :param client: the AitoClient instance
    :type client: AitoClient
    :param table_name: the name of the table
    :type table_name: str
    """
    client.request(request_obj=aito_requests.DeleteTableSchemaRequest(table_name=table_name))
    LOG.info(f'table `{table_name}` deleted')


def create_column(client: AitoClient, table_name: str, column_name: str, schema: Dict):
    """`add or replace a column <https://aito.ai/docs/api/#put-api-v1-schema-table-column>`__

    .. note::

        requires the client to be setup with the READ-WRITE API key

    :param client: the AitoClient instance
    :type client: AitoClient
    :param table_name: the name of the table containing the column
    :type table_name: str
    :param column_name: the name of the column
    :type column_name: str
    :param schema: the schema of the column
    :type schema: Dict
    """
    client.request(request_obj=aito_requests.CreateColumnSchemaRequest(
        table_name=table_name, column_name=column_name, schema=schema)
    )
    LOG.info(f'column `{table_name}.{column_name}` created')


def get_column_schema(client: AitoClient, table_name: str, column_name: str) -> AitoColumnTypeSchema:
    """`get the schema of the specified column <https://aito.ai/docs/api/#get-api-v1-schema-table-column>`__

    :param client: the AitoClient instance
    :type client: AitoClient
    :param table_name: the name of the table containing the column
    :type table_name: str
    :param column_name: the name of the column
    :type column_name: str
    :return: the column schema
    :rtype: AitoColumnTypeSchema
    """
    resp = client.request(request_obj=aito_requests.GetColumnSchemaRequest(table_name=table_name, column_name=column_name))
    return resp.schema


def delete_column(client: AitoClient, table_name: str, column_name: str):
    """`delete a column of a table <https://aito.ai/docs/api/#delete-api-v1-schema-column>`__

    .. note::

        requires the client to be setup with the READ-WRITE API key

    :param client: the AitoClient instance
    :type client: AitoClient
    :param table_name: the name of the table containing the column
    :type table_name: str
    :param column_name: the name of the column
    :type column_name: str
    """
    client.request(request_obj=aito_requests.DeleteColumnSchemaRequest(table_name=table_name, column_name=column_name))
    LOG.info(f'column `{table_name}.{column_name}` deleted')


def get_existing_tables(client: AitoClient) -> List[str]:
    """get a list of existing tables in the instance

    :param client: the AitoClient instance
    :type client: AitoClient
    :return: list of the names of existing tables
    :rtype: List[str]
    """
    db_schema = get_database_schema(client)
    return db_schema.tables


def check_table_exists(client: AitoClient, table_name: str) -> bool:
    """check if a table exists in the instance

    :param client: the AitoClient instance
    :type client: AitoClient
    :param table_name: the name of the table
    :type table_name: str
    :return: True if the table exists
    :rtype: bool
    """
    existing_tables = get_existing_tables(client)
    return table_name in existing_tables


def rename_table(client: AitoClient, old_name: str, new_name: str, replace: bool = False):
    """`rename a table <https://aito.ai/docs/api/#post-api-v1-schema-rename>`__

    .. note::

        requires the client to be setup with the READ-WRITE API key

    :param client: the AitoClient instance
    :type client: AitoClient
    :param old_name: the name of the table to be renamed
    :type old_name: str
    :param new_name: the new name of the table
    :type new_name: str
    :param replace: replace an existing table of which name is the new name, defaults to False
    :type replace: bool, optional
    """
    client.request(
        method='POST',
        endpoint='/api/v1/schema/_rename',
        query={'from': old_name, 'rename': new_name, 'replace': replace}
    )


def copy_table(client: AitoClient, table_name: str, copy_table_name: str, replace: bool = False):
    """`copy a table <https://aito.ai/docs/api/#post-api-v1-schema-copy>`__

    .. note::

        requires the client to be setup with the READ-WRITE API key

    :param client: the AitoClient instance
    :type client: AitoClient
    :param table_name: the name of the table to be copied
    :type table_name: str
    :param copy_table_name: the name of the new copy table
    :type copy_table_name: str
    :param replace: replace an existing table of which name is the name of the copy table, defaults to False
    :type replace: bool, optional
    """
    client.request(
        method='POST',
        endpoint='/api/v1/schema/_copy',
        query={'from': table_name, 'copy': copy_table_name, 'replace': replace}
    )


def optimize_table(client: AitoClient, table_name):
    """`optimize <https://aito.ai/docs/api/#post-api-v1-data-table-optimize>`__
    the specified table after uploading the data

    :param client: the AitoClient instance
    :type client: AitoClient
    :param table_name: the name of the table
    :type table_name: str
    :return:
    :rtype:
    """
    try:
        client.request(method='POST', endpoint=f'/api/v1/data/{table_name}/optimize', query={})
    except Exception as e:
        LOG.error(f'failed to optimize: {e}')
    LOG.info(f'table {table_name} optimized')


def delete_entries(client: AitoClient, query: Dict):
    """`Delete the entries <https://aito.ai/docs/api/#post-api-v1-data-delete>`__ according to the criteria
    given in the query

    :param client: the AitoClient instance
    :type client: AitoClient
    :param query: the query to describe the target table and filters for which entries to delete.
    :type query: Dict
    """
    client.request(request_obj=aito_requests.DeleteEntriesRequest(query=query))
    LOG.info(f'entries deleted')


def upload_entries(
        client: AitoClient,
        table_name: str,
        entries: Iterable[Dict],
        batch_size: int = 1000,
        optimize_on_finished: bool = True
):
    """populate table entries by batches of batch_size

    .. note::

        requires the client to be setup with the READ-WRITE API key

    :param client: the AitoClient instance
    :type client: AitoClient
    :param table_name: the name of the table
    :type table_name: str
    :param entries: iterable of the table entries
    :type entries: Iterable[Dict]
    :param batch_size: the batch size, defaults to 1000
    :type batch_size: int, optional
    :param optimize_on_finished: `optimize <https://aito.ai/docs/api/#post-api-v1-data-table-optimize>`__
        the table on finished, defaults to True
    :type optimize_on_finished: bool

    Upload a Pandas DataFrame

    >>> import pandas as pd
    >>> df = pd.DataFrame({'height': [94, 170], 'weight': [31, 115], 'depth':  [ 3,  29]})
    >>> entries = df.to_dict(orient='records')
    >>> client.upload_entries(table_name='specifications', entries=entries) # doctest: +SKIP

    Upload a generator of entries

    >>> def entries_generator(start, end):
    ...     for idx in range(start,clea end):
    ...         entry = {'id': idx}
    ...         yield entry
    >>> client.upload_entries(
    ...     table_name="table_name",
    ...     entries=entries_generator(start=0, end=10000),
    ...     batch_size=500,
    ...     optimize_on_finished=False
    ... ) # doctest: +SKIP
    """

    LOG.info(f'uploading entries to table `{table_name}` with batch size of {batch_size}...')
    begin_index, last_index, populated_count = 0, 0, 0
    entries_batch = []

    def _upload_single_batch(begin_idx, last_idx, populated_c, batch_content):
        try:
            LOG.debug(f'uploading batch {begin_idx}:{last_idx}...')
            client.request(request_obj=aito_requests.UploadEntriesRequest(table_name=table_name, entries=batch_content))
            populated_c += len(batch_content)
            LOG.info(f'uploaded batch {begin_idx}:{last_idx}')
        except Exception as e:
            LOG.error(f'batch {begin_idx}:{last_idx} failed: {e}')
        return populated_c

    for entry in entries:
        entries_batch.append(entry)
        last_index += 1

        if last_index % batch_size == 0:
            populated_count = _upload_single_batch(begin_index, last_index, populated_count, entries_batch)

            begin_index = last_index
            entries_batch = []

    # upload the remaining entries
    if last_index % batch_size != 0:
        populated_count = _upload_single_batch(begin_index, last_index, populated_count, entries_batch)

    if populated_count == 0:
        raise Exception("failed to upload any data into Aito")

    LOG.info(f'uploaded {populated_count}/{last_index} entries to table `{table_name}`')

    if optimize_on_finished:
        optimize_table(client, table_name)


def initiate_upload_file(client: AitoClient, table_name: str) -> Dict:
    """`Initial uploading a file to a table <https://aito.ai/docs/api/#post-api-v1-data-table-file>`__

    :param client: the AitoClient instance
    :type client: AitoClient
    :param table_name: the name of the table to be uploaded
    :type table_name: str
    :return: The details to execute the S3 upload and the upload session's id
    :rtype: Dict
    """
    LOG.debug('initiating file upload...')
    r = client.request(request_obj=aito_requests.InitiateFileUploadRequest(table_name=table_name))
    return r.json


def upload_binary_file_to_s3(initiate_upload_file_response: Dict, binary_file: BinaryIO):
    """Upload a binary file to AWS S3 with using the information from :func:`.initiate_upload_file`

    :param initiate_upload_file_response: the result from :func:`.initiate_upload_file`
    :type initiate_upload_file_response: Dict
    :param binary_file: binary file object
    :type binary_file: BinaryIO
    """
    LOG.debug('uploading binary file to S3...')
    LOG.debug('getting session id and upload url...')
    s3_url = initiate_upload_file_response['url']
    upload_req_method = initiate_upload_file_response['method']
    LOG.debug('uploading file file to S3...')
    try:
        r = requestslib.request(upload_req_method, s3_url, data=binary_file)
        r.raise_for_status()
    except Exception as e:
        raise Exception(f'failed to upload file to S3: {e}')
    LOG.debug('uploaded file to S3')


def trigger_file_processing(client: AitoClient, table_name: str, session_id: str):
    """`Trigger file processing of uploading a file to a table
    <https://aito.ai/docs/api/#post-api-v1-data-table-file-uuid>`__

    :param client: the AitoClient instance
    :type client: AitoClient
    :param table_name: the name of the table to be uploaded
    :type table_name: str
    :param session_id: The upload session id from :func:`.initiate_upload_file`
    :type session_id: str
    """
    LOG.debug('triggering file processing...')
    client.request(request_obj=aito_requests.TriggerFileProcessingRequest(table_name=table_name, session_id=session_id))
    LOG.info('triggered file processing')


def poll_file_processing_status(client: AitoClient, table_name: str, session_id: str, polling_time: int = 10):
    """Polling the `file processing status <https://aito.ai/docs/api/#get-api-v1-data-table-file-uuid>`__ until
    the processing finished

    :param client: the AitoClient instance
    :type client: AitoClient
    :param table_name: the name of the table to be uploaded
    :type table_name: str
    :param session_id: The upload session id from :func:`.initiate_upload_file`
    :type session_id: str
    :param polling_time: polling wait time
    :type polling_time: int
    """
    LOG.debug('polling processing status...')
    while True:
        try:
            processing_progress_resp = client.request(
                request_obj=aito_requests.GetFileProcessingRequest(table_name=table_name, session_id=session_id)
            )
            status = processing_progress_resp['status']
            LOG.debug(f"completed count: {status['completedCount']}, throughput: {status['throughput']}")
            if processing_progress_resp['errors']['message'] != 'Last 0 failing rows':
                LOG.error(processing_progress_resp['errors'])
            if status['finished']:
                break
        except Exception as e:
            LOG.debug(f'failed to get file upload status: {e}')
        time.sleep(polling_time)


def upload_binary_file(
        client: AitoClient,
        table_name: str,
        binary_file: BinaryIO,
        polling_time: int = 10,
        optimize_on_finished: bool = True
):
    """`upload a binary file object to a table <https://aito.ai/docs/api/#post-api-v1-data-table-file>`__

    .. note::

        requires the client to be setup with the READ-WRITE API key

    :param client: the AitoClient instance
    :type client: AitoClient
    :param table_name: the name of the table to be uploaded
    :type table_name: str
    :param binary_file: binary file object
    :type binary_file: BinaryIO
    :param polling_time: polling wait time
    :type polling_time: int
    :param optimize_on_finished: :func:`optimize_table` when finished uploading, defaults to True
    :type optimize_on_finished: bool
    """
    LOG.debug(f'uploading file object to table `{table_name}`...')
    init_upload_resp = initiate_upload_file(client=client, table_name=table_name)
    upload_binary_file_to_s3(initiate_upload_file_response=init_upload_resp, binary_file=binary_file)
    upload_session_id = init_upload_resp['id']
    trigger_file_processing(client=client, table_name=table_name, session_id=upload_session_id)
    poll_file_processing_status(
        client=client, table_name=table_name, session_id=upload_session_id, polling_time=polling_time
    )

    LOG.info(f'uploaded file object to table `{table_name}`')
    if optimize_on_finished:
        optimize_table(client, table_name)


def upload_file(
        client: AitoClient,
        table_name: str,
        file_path: PathLike,
        polling_time: int = 10,
        optimize_on_finished: bool = True
):
    """`upload a file <https://aito.ai/docs/api/#post-api-v1-data-table-file>`__ to the specfied table

    .. note::

        requires the client to be setup with the READ-WRITE API key

    :param client: the AitoClient instance
    :type client: AitoClient
    :param table_name: the name of the table
    :type table_name: str
    :param file_path: path to the file to be uploaded
    :type file_path: PathLike
    :param polling_time: polling wait time
    :type polling_time: int
    :param optimize_on_finished: :func:`optimize_table` when finished uploading, defaults to True
    :type optimize_on_finished: bool
    :raises ValueError: incorrect file extension, should be .ndjson.gz
    """
    if not check_file_is_gzipped(file_path):
        raise ValueError(f'file {file_path} is not a gzip-compressed ndjson file')
    with open(file_path, 'rb') as f:
        upload_binary_file(
            client=client,
            table_name=table_name,
            binary_file=f,
            polling_time=polling_time,
            optimize_on_finished=optimize_on_finished
        )


def quick_add_table(
        client: AitoClient, input_file: Union[Path, PathLike], table_name: str = None, input_format: str = None
):
    """Create a table and upload a file to the table, using the default inferred schema

    :param client: the AitoClient instance
    :type client: AitoClient
    :param input_file: path to the input file to be uploaded
    :type input_file: Union[Path, PathLike]
    :param table_name: the name of the table, defaults to the name of the input file
    :type table_name: Optional[str]
    :param input_format: specify the format of the input file, defaults to the input file extension
    :type input_format: Optional[str]
    """
    df_handler = DataFrameHandler()

    try:
        in_f_path = Path(input_file)
    except Exception:
        raise ValueError(f'invalid path: {input_file}')
    in_format = in_f_path.suffixes[0].replace('.', '') if input_format is None else input_format
    if in_format not in df_handler.allowed_format:
        raise ValueError(f'invalid file format {in_format}. Must be one of {"|".join(df_handler.allowed_format)}')

    table_name = in_f_path.stem if table_name is None else table_name

    converted_tmp_file = tempfile.NamedTemporaryFile(mode='w', suffix='.ndjson.gz', delete=False)

    converted_df = df_handler.convert_file(
        read_input=in_f_path,
        write_output=converted_tmp_file.name,
        in_format=in_format,
        out_format='ndjson',
        convert_options={'compression': 'gzip'}
    )
    converted_tmp_file.close()

    inferred_schema = AitoTableSchema.infer_from_pandas_data_frame(converted_df)
    create_table(client, table_name, inferred_schema)

    with open(converted_tmp_file.name, 'rb') as in_f:
        upload_binary_file(client=client, table_name=table_name, binary_file=in_f)
    converted_tmp_file.close()
    unlink(converted_tmp_file.name)


def _create_job_request(
        job_endpoint: Optional[str] = None,
        query: Optional[Union[List, Dict]] = None,
        request_obj: Optional[Union[aito_requests.QueryAPIRequest, aito_requests.CreateJobRequest]] = None
) -> aito_requests.CreateJobRequest:
    """Return the Create Job Request"""
    if request_obj is None:
        if job_endpoint is not None and query is not None:
            create_job_req = aito_requests.CreateJobRequest(endpoint=job_endpoint, query=query)
        else:
            raise TypeError("create_job() requires either 'request_obj' or 'job_endpoint' and 'query'")
    else:
        if isinstance(request_obj, aito_requests.QueryAPIRequest):
            create_job_req = aito_requests.CreateJobRequest.from_query_api_request(request_obj=request_obj)
        else:
            create_job_req = request_obj
    return create_job_req


def create_job(
        client: AitoClient,
        job_endpoint: Optional[str] = None,
        query: Optional[Union[List, Dict]] = None,
        request_obj: Optional[Union[aito_requests.QueryAPIRequest, aito_requests.CreateJobRequest]] = None,
) -> aito_responses.CreateJobResponse:
    """Create a `job <https://aito.ai/docs/api/#post-api-v1-jobs-query>`__
    for a query that takes longer than 30 seconds to run

    :param client: the AitoClient instance
    :type client: AitoClient
    :param job_endpoint: the job endpoint
    :type job_endpoint: Optional[str]
    :param query: the query for the endpoint
    :type query: Optional[Union[List, Dict]]
    :param request_obj: a :class:`.CreateJobRequest` or an :class:`.QueryAPIRequest` instance
    :type request_obj: Optional[Union[aito_requests.QueryAPIRequest, aito_requests.CreateJobRequest]]
    :return: job information
    :rtype: Dict
    """
    create_job_req = _create_job_request(job_endpoint=job_endpoint, query=query, request_obj=request_obj)
    resp = client.request(request_obj=create_job_req)
    return resp


def get_job_status(client: AitoClient, job_id: str) -> aito_responses.GetJobStatusResponse:
    """`Get the status of a job <https://aito.ai/docs/api/#get-api-v1-jobs-uuid>`__ with the specified job id

    :param client: the AitoClient instance
    :type client: AitoClient
    :param job_id: the id of the job session
    :type job_id: str
    :return: job status
    :rtype: Dict
    """
    resp = client.request(request_obj=aito_requests.GetJobStatusRequest(job_id=job_id))
    return resp


def get_job_result(client: AitoClient, job_id: str) -> aito_responses.BaseResponse:
    """`Get the result of a job <https://aito.ai/docs/api/#get-api-v1-jobs-uuid-result>`__ with the specified job id

    :param client: the AitoClient instance
    :type client: AitoClient
    :param job_id: the id of the job
    :type job_id: str
    :return: the job result
    :rtype: Dict
    """
    resp = client.request(request_obj=aito_requests.GetJobResultRequest(job_id=job_id))
    return resp


def job_request(
        client: AitoClient,
        job_endpoint: Optional[str] = None,
        query: Optional[Union[List, Dict]] = None,
        request_obj: Optional[Union[aito_requests.QueryAPIRequest, aito_requests.CreateJobRequest]] = None,
        polling_time: int = 10
) -> aito_responses.BaseResponse:
    """make a request to an Aito API endpoint using `Job <https://aito.ai/docs/api/#post-api-v1-jobs-query>`__

    This method should be used for requests that take longer than 30 seconds, e.g: evaluate

    The following query evaluate the performance of a predict query that uses the name of
    a product to predict its category

    >>> response = client.job_request(
    ...     job_endpoint='/api/v1/jobs/_evaluate',
    ...     query={
    ...         "test": { "$index": { "$mod": [4, 0] } },
    ...         "evaluate": {
    ...             "from": "products",
    ...             "where": { "name": { "$get": "name" } },
    ...             "predict": "category"
    ...         }
    ...     }
    ... )
    >>> print(response["accuracy"]) # doctest: +ELLIPSIS
    0.72...

    :param client: the AitoClient instance
    :type client: AitoClient
    :param job_endpoint: job end point
    :type job_endpoint: Optional[str]
    :param query: the query for the endpoint
    :type query: Optional[Union[List, Dict]]
    :param request_obj: a :class:`.CreateJobRequest` or an :class:`.QueryAPIRequest` instance
    :type request_obj: Optional[Union[aito_requests.QueryAPIRequest, aito_requests.CreateJobRequest]]
    :param polling_time: polling wait time, defaults to 10
    :type polling_time: int
    :raises RequestError: an error occurred during the execution of the job
    :return: request JSON content
    :rtype: Dict
    """
    create_job_req = _create_job_request(job_endpoint=job_endpoint, query=query, request_obj=request_obj)
    create_job_resp = client.request(request_obj=create_job_req)
    job_id = create_job_resp.id
    LOG.debug(f'polling job {job_id} status...')
    while True:
        job_status_resp = get_job_status(client, job_id)
        if job_status_resp.finished:
            LOG.debug(f'job {job_id} finished')
            break
        time.sleep(polling_time)
    job_result_resp = get_job_result(client, job_id)
    casted_job_result_resp = create_job_req.result_response_cls(job_result_resp.json)
    return casted_job_result_resp


def get_table_size(client: AitoClient, table_name: str) -> int:
    """return the number of entries of the specified table

    :param client: the AitoClient instance
    :type client: AitoClient
    :param table_name: the name of the table
    :type table_name: str
    :return: the number of entries in the table
    :rtype: int
    """
    resp = generic_query(client=client, query={'from': table_name})
    return resp['total']


def query_entries(
        client: AitoClient, table_name: str, offset: int = 0, limit: int = 10, select: List[str] = None
) -> List[Dict]:
    """`query <https://aito.ai/docs/api/#post-api-v1-query>`__ entries of the specified table

    use offset and limit for `pagination <https://aito.ai/docs/api/#pagination>`__

    :param client: the AitoClient instance
    :type client: AitoClient
    :param table_name: the name of the table
    :type table_name: str
    :param offset: the offset of the first entry, defaults to 0
    :type offset: int, optional
    :param limit: the number of entries to be returned, defaults to 10
    :type limit: int, optional
    :param select: specify the fields of the entry, including link fields, to be returned
    :type select: List[str], optional
    :return: the table entries
    :rtype: List[Dict]
    """
    resp = generic_query(client=client, query={'from': table_name, 'offset': offset, 'limit': limit, 'select': select})
    return resp['hits']


def query_all_entries(
        client: AitoClient,
        table_name: str,
        select: List[str] = None,
        batch_size: int = 5000
) -> List[Dict]:
    """`query <https://aito.ai/docs/api/#post-api-v1-query>`__  all entries of the specified table

    :param client: the AitoClient instance
    :type client: AitoClient
    :param table_name: the name of the table
    :type table_name: str
    :param select: specify the fields of the entry, including link fields, to be returned
    :type select: List[str], optional
    :param batch_size: the number of entries to be queried at once, defaults to 5000
    :type batch_size: int, optional
    :return: list of all entries in the table
    :rtype: List[Dict]
    """
    LOG.debug(f'getting all entries of table {table_name} in batch of {batch_size}...')
    table_size = get_table_size(client, table_name)
    all_entries = []
    begin_idx = 0
    while begin_idx < table_size:
        last_idx = begin_idx + batch_size if begin_idx + batch_size <= table_size else table_size
        LOG.debug(f'getting table chunk {begin_idx}:{last_idx}...')
        all_entries += query_entries(client, table_name, offset=begin_idx, limit=batch_size, select=select)
        LOG.debug(f'queried chunk {begin_idx}:{last_idx}')
        begin_idx += batch_size
    LOG.info(f'queried all entries of table `{table_name}`')
    return all_entries


def download_table(
        client: AitoClient,
        table_name: str,
        output_folder: PathLike,
        file_name: str = None,
        batch_size: int = 5000,
        gzip_output: bool = False
):
    """download a table to a NDJSON file or a gzipped NDJSON file

    :param client: the AitoClient instance
    :type client: AitoClient
    :param table_name: the name of the table
    :type table_name: str
    :param output_folder: the folder where the output file is written to
    :type output_folder: PathLike
    :parm file_name: the name of the output file, defaults to None in which the table name is used the file name
    :type file_name: str
    :param batch_size: the number of entries to be downloaded at once, defaults to 5000
    :type batch_size: int
    :param gzip_output: gzip the output file, defaults to False
    :type gzip_output: bool
    """
    if not file_name:
        file_name = table_name
    out_file_path = Path(output_folder) / f'{file_name}.ndjson'
    if out_file_path.exists():
        LOG.warning(f'output file {out_file_path} already exists')
    LOG.debug(f'downloading table `{table_name}` to {out_file_path}')
    table_size = get_table_size(client, table_name)
    begin_idx = 0
    while begin_idx < table_size:
        last_idx = begin_idx + batch_size if begin_idx + batch_size <= table_size else table_size
        LOG.debug(f'downloading table chunk {begin_idx}:{last_idx}...')
        entries_batch = query_entries(client=client, table_name=table_name, offset=begin_idx, limit=batch_size)
        with out_file_path.open('a+') as f:
            ndjson.dump(entries_batch, f)
            if last_idx != table_size:
                f.write('\n')
        LOG.debug(f'downloaded table chunk {begin_idx}:{last_idx}')
        begin_idx += batch_size
    if gzip_output:
        gzip_file(out_file_path, keep=False)
    LOG.info(f'downloaded table `{table_name}` to {out_file_path}')


def quick_predict_and_evaluate(
        client: AitoClient,
        from_table: str,
        predicting_field: str
) -> Tuple[Dict, Dict]:
    """generate an example predict query to predict a field and the corresponding evaluate query

    The example query will use all fields of the table as the hypothesis and the first entry of the table as the
    input data

    :param client: the AitoClient instance
    :type client: AitoClient
    :param from_table: the name of the table the will be use as context for prediction.
    :type from_table: str
    :param predicting_field: the name of the predicting field. If the field belongs to a linked table,
        it should be in the format of <column_with_link>.<field_name>
    :type predicting_field: str
    :return: a tuple contains the predict query and the prediction result
    :rtype: Tuple[Dict, Dict]
    """
    database_schema = get_database_schema(client)
    table_schema = database_schema[from_table]

    predicting_field_splitted = predicting_field.split('.')
    is_predicting_a_linked_field = len(predicting_field_splitted) == 2

    predicting_col = predicting_field if not is_predicting_a_linked_field else predicting_field_splitted[0]
    if predicting_col not in table_schema.columns:
        raise ValueError(f"table `{from_table}` does not have column `{predicting_col}`")
    if is_predicting_a_linked_field:
        linked_col = predicting_field_splitted[1]
        linked_table = table_schema.links[predicting_col].table_name
        if linked_col not in database_schema[linked_table].columns:
            raise ValueError(f"linked table `{linked_table}` does not have column `{linked_col}`")

    table_first_entry_res = query_entries(client, from_table, limit=1)
    if not table_first_entry_res:
        raise ValueError(f"table `{from_table}` is empty, cannot generate example query")
    table_first_entry = table_first_entry_res[0]

    predict_query = {
        'from': from_table,
        'where': {
            col: table_first_entry.get(col) for col in table_schema.columns if col != predicting_col
        },
        'predict': predicting_field,
        'select': ['$p', 'feature', '$why']
    }

    evaluating_pred_query = dict(predict_query)
    for col in evaluating_pred_query['where']:
        evaluating_pred_query['where'][col] = {'$get': col}
    evaluating_pred_query.pop('select')

    evaluate_query = {
        'test': {'$index': {"$mod": [10, 0]}},
        'evaluate': evaluating_pred_query
    }

    return predict_query, evaluate_query


def quick_predict(
        client: AitoClient,
        from_table: str,
        predicting_field: str
):
    """generate an example predict query to predict a field

    The example query will use all fields of the table as the hypothesis and the first entry of the table as the
    input data

    :param client: the AitoClient instance
    :type client: AitoClient
    :param from_table: the name of the table the will be use as context for prediction.
    :type from_table: str
    :param predicting_field: the name of the predicting field. If the field belongs to a linked table,
        it should be in the format of <column_with_link>.<field_name>
    :type predicting_field: str
    :return: The example predict query
    :rtype: Dict
    """
    predict_query, evaluate_query = quick_predict_and_evaluate(
        client=client, from_table=from_table, predicting_field=predicting_field
    )
    return predict_query


def search(
        client: AitoClient, query: Dict, raise_for_status: Optional[bool] = None, use_job: bool = False
) -> Union[aito_responses.SearchResponse, RequestError]:
    """send a query to the `Search API <https://aito.ai/docs/api/#post-api-v1-search>`__

    :param client: the AitoClient instance
    :type client: AitoClient
    :param query: the search query
    :type query: Dict
    :param raise_for_status: raise :class:`.RequestError` if the request fails instead of returning the error
        If set to None, value from Client will be used. Defaults to None
    :type raise_for_status: Optional[bool]
    :param use_job: use job fo request that takes longer than 30 seconds, defaults to False
    :type use_job: bool
    :return: :class:`SearchResponse` or :class:`.RequestError` if an error occurred and not raise_for_status
    :rtype: SearchResponse
    """
    req = aito_requests.SearchRequest(query=query)
    if use_job:
        return job_request(client=client, request_obj=req)
    return client.request(request_obj=req, raise_for_status=raise_for_status)


def predict(
        client: AitoClient, query: Dict, raise_for_status: Optional[bool] = None, use_job: bool = False
) -> Union[aito_responses.PredictResponse, RequestError]:
    """send a query to the `Predict API <https://aito.ai/docs/api/#post-api-v1-predict>`__

    :param client: the AitoClient instance
    :type client: AitoClient
    :param query: the predict query
    :type query: Dict
    :param raise_for_status: raise :class:`.RequestError` if the request fails instead of returning the error
        If set to None, value from Client will be used. Defaults to None
    :type raise_for_status: Optional[bool]
    :param use_job: use job fo request that takes longer than 30 seconds, defaults to False
    :type use_job: bool
    :return: :class:`.PredictResponse` or :class:`.RequestError` if an error occurred and not raise_for_status
    :rtype: Union[PredictResponse, RequestError]
    """
    req = aito_requests.PredictRequest(query)
    if use_job:
        return job_request(client=client, request_obj=req)
    return client.request(request_obj=req, raise_for_status=raise_for_status)


def recommend(
        client: AitoClient, query: Dict, raise_for_status: Optional[bool] = None, use_job: bool = False
) -> Union[aito_responses.RecommendResponse, RequestError]:
    """send a query to the `Recommend API <https://aito.ai/docs/api/#post-api-v1-recommend>`__

    :param client: the AitoClient instance
    :type client: AitoClient
    :param query: the recommend query
    :type query: Dict
    :param raise_for_status: raise :class:`.RequestError` if the request fails instead of returning the error
        If set to None, value from Client will be used. Defaults to None
    :type raise_for_status: Optional[bool]
    :param use_job: use job fo request that takes longer than 30 seconds, defaults to False
    :type use_job: bool
    :return: :class:`.RecommendResponse`  or :class:`.RequestError` if an error occurred and not raise_for_status
    :rtype: Union[RecommendResponse, RequestError]
    """
    req = aito_requests.RecommendRequest(query)
    if use_job:
        return job_request(client=client, request_obj=req)
    return client.request(request_obj=req, raise_for_status=raise_for_status)


def evaluate(
        client: AitoClient, query: Dict, raise_for_status: Optional[bool] = None, use_job: bool = False
) -> Union[aito_responses.EvaluateResponse, RequestError]:
    """send a query to the `Evaluate API <https://aito.ai/docs/api/#post-api-v1-evaluate>`__

    :param client: the AitoClient instance
    :type client: AitoClient
    :param query: the evaluate query
    :type query: Dict
    :param raise_for_status: raise :class:`.RequestError` if the request fails instead of returning the error
        If set to None, value from Client will be used. Defaults to None
    :type raise_for_status: Optional[bool]
    :param use_job: use job fo request that takes longer than 30 seconds, defaults to False
    :type use_job: bool
    :return: :class:`.EvaluateResponse` or :class:`.RequestError` if an error occurred and not raise_for_status
    :rtype: Union[EvaluateResponse, RequestError]
    """
    req = aito_requests.EvaluateRequest(query)
    if use_job:
        return job_request(client=client, request_obj=req)
    return client.request(request_obj=req, raise_for_status=raise_for_status)


def similarity(
        client: AitoClient, query: Dict, raise_for_status: Optional[bool] = None, use_job: bool = False
) -> Union[aito_responses.SimilarityResponse, RequestError]:
    """send a query to the `Similarity API <https://aito.ai/docs/api/#post-api-v1-similarity>`__

    :param client: the AitoClient instance
    :type client: AitoClient
    :param query: the similarity query
    :type query: Dict
    :param raise_for_status: raise :class:`.RequestError` if the request fails instead of returning the error
        If set to None, value from Client will be used. Defaults to None
    :type raise_for_status: Optional[bool]
    :param use_job: use job fo request that takes longer than 30 seconds, defaults to False
    :type use_job: bool
    :return: :class:`.SimilarityResponse` or :class:`.RequestError` if an error occurred and not raise_for_status
    :rtype: Union[SimilarityResponse, RequestError]
    """
    req = aito_requests.SimilarityRequest(query)
    if use_job:
        return job_request(client=client, request_obj=req)
    return client.request(request_obj=req, raise_for_status=raise_for_status)


def match(
        client: AitoClient, query: Dict, raise_for_status: Optional[bool] = None, use_job: bool = False
) -> Union[aito_responses.MatchResponse, RequestError]:
    """send a query to the `Match API <https://aito.ai/docs/api/#post-api-v1-match>`__

    :param client: the AitoClient instance
    :type client: AitoClient
    :param query: the match query
    :type query: Dict
    :param raise_for_status: raise :class:`.RequestError` if the request fails instead of returning the error
        If set to None, value from Client will be used. Defaults to None
    :type raise_for_status: Optional[bool]
    :param use_job: use job fo request that takes longer than 30 seconds, defaults to False
    :type use_job: bool
    :return: :class:`.MatchResponse` or :class:`.RequestError` if an error occurred and not raise_for_status
    :rtype: Union[MatchResponse, RequestError]
    """
    req = aito_requests.MatchRequest(query)
    if use_job:
        return job_request(client=client, request_obj=req)
    return client.request(request_obj=req, raise_for_status=raise_for_status)


def relate(
        client: AitoClient, query: Dict, raise_for_status: Optional[bool] = None, use_job: bool = False
) -> Union[aito_responses.RelateResponse, RequestError]:
    """send a query to the `Relate API <https://aito.ai/docs/api/#post-api-v1-relate>`__

    :param client: the AitoClient instance
    :type client: AitoClient
    :param query: the relate query
    :type query: Dict
    :param raise_for_status: raise :class:`.RequestError` if the request fails instead of returning the error
        If set to None, value from Client will be used. Defaults to None
    :type raise_for_status: Optional[bool]
    :param use_job: use job fo request that takes longer than 30 seconds, defaults to False
    :type use_job: bool
    :return: :class:`.RelateResponse` or :class:`.RequestError` if an error occurred and not raise_for_status
    :rtype: Union[RelateResponse, RequestError]
    """
    req = aito_requests.RelateRequest(query)
    if use_job:
        return job_request(client=client, request_obj=req)
    return client.request(request_obj=req, raise_for_status=raise_for_status)


def generic_query(
        client: AitoClient, query: Dict, raise_for_status: Optional[bool] = None, use_job: bool = False
) -> Union[aito_responses.HitsResponse, RequestError]:
    """send a query to the `Generic Query API <https://aito.ai/docs/api/#post-api-v1-qyert>`__

    :param client: the AitoClient instance
    :type client: AitoClient
    :param query: the query
    :type query: Dict
    :param raise_for_status: raise :class:`.RequestError` if the request fails instead of returning the error
        If set to None, value from Client will be used. Defaults to None
    :type raise_for_status: Optional[bool]
    :param use_job: use job fo request that takes longer than 30 seconds, defaults to False
    :type use_job: bool
    :return: :class:`HitsResponse` or :class:`.RequestError` if an error occurred and not raise_for_status
    :rtype: Union[HitsResponse, RequestError]
    """
    req = aito_requests.GenericQueryRequest(query)
    if use_job:
        return job_request(client=client, request_obj=req)
    return client.request(request_obj=req, raise_for_status=raise_for_status)



import json
import sys
from typing import Dict, List

from aito.schema import AitoTableSchema
from aito.utils.data_frame_handler import DataFrameHandler
from .sub_command import SubCommand
from ..parser import PathArgType, InputArgType, ParseError, try_load_json


class ConvertFromFormatSubCommand(SubCommand):
    def build_parser(self, parser):
        # add share arguments between formats
        either_use_or_create_schema = parser.add_mutually_exclusive_group()
        either_use_or_create_schema.add_argument(
            '-c', '--create-table-schema', metavar='schema-output-file', type=PathArgType(parent_must_exist=True),
            help='create an inferred aito schema and write to output file'
        )
        either_use_or_create_schema.add_argument(
            '-s', '--use-table-schema', metavar='schema-input-file', type=PathArgType(must_exist=True),
            help='convert the data to match the input table schema'
        )
        parser.add_argument('-j', '--json', action='store_true', help='convert to json format')
        parser.add_argument(
            'input', default='-', type=InputArgType(), nargs='?',
            help="path to the input file (when no input file is given or when input is -, read from the standard input)"
        )

    @staticmethod
    def parsed_args_to_data_frame_handler_convert_args(parsed_args: Dict) -> Dict:
        in_format = parsed_args['input-format']
        convert_args = {
            'read_input': parsed_args['input'],
            'write_output': sys.stdout,
            'in_format': parsed_args['input-format'],
            'out_format': 'json' if parsed_args['json'] else 'ndjson',
            'read_options': {},
            'convert_options': {},
        }

        if parsed_args['use_table_schema']:
            with parsed_args['use_table_schema'].open() as f:
                table_schema = try_load_json(f, 'table schema')
            convert_args['use_table_schema'] = table_schema

        if in_format == 'csv':
            convert_args['read_options']['delimiter'] = parsed_args['delimiter']
            convert_args['read_options']['decimal'] = parsed_args['decimal']

        if in_format == 'excel':
            if parsed_args['input'] == sys.stdin:
                raise ParseError('input must be a file path for excel files')
            if parsed_args['one_sheet']:
                convert_args['read_options']['sheet_name'] = parsed_args['one_sheet']
        return convert_args

    def parse_and_execute(self, parsed_args: Dict):
        parsed_convert_args = self.parsed_args_to_data_frame_handler_convert_args(parsed_args)
        output_schema_path = parsed_args['create_table_schema'] if parsed_args['create_table_schema'] else None

        converted_df = DataFrameHandler().convert_file(**parsed_convert_args)
        if output_schema_path:
            inferred_schema = AitoTableSchema.infer_from_pandas_data_frame(converted_df)
            with output_schema_path.open(mode='w') as f:
                json.dump(inferred_schema.to_json_serializable(), f, indent=2, sort_keys=True)
        return 0


class ConvertFromCSVSubCommand(ConvertFromFormatSubCommand):
    def __init__(self):
        super().__init__('csv', 'convert CSV data')

    def build_parser(self, parser):
        super().build_parser(parser)
        parser.add_csv_format_default_arguments()


class ConvertFromExcelSubCommand(ConvertFromFormatSubCommand):
    def __init__(self):
        super().__init__('excel', 'convert EXCEL data')

    def build_parser(self, parser):
        super().build_parser(parser)
        parser.add_excel_format_default_arguments()
        parser.description = 'Convert EXCEL data, accept both xls and xlsx'


class ConvertSubCommand(SubCommand):
    _default_sub_commands = [
        ConvertFromCSVSubCommand(),
        ConvertFromExcelSubCommand(),
        ConvertFromFormatSubCommand('json', 'convert JSON data'),
        ConvertFromFormatSubCommand('ndjson', 'convert NDJSON data'),
    ]

    def __init__(self, sub_commands: List[SubCommand] = None):
        super().__init__('convert', 'convert from a given format into NDJSON|JSON')
        if not sub_commands:
            sub_commands = self._default_sub_commands
        self._sub_commands_map = {cmd.name: cmd for cmd in sub_commands}

    def build_parser(self, parser):
        parser.epilog = '''To see help for a specific format:
  aito convert <input-format> - h

When no input or when input is -, read standard input.
You must use input file instead of standard input for excel file
'''
        sub_commands_subparsers = parser.add_subparsers(
            title='input-format',
            dest='input-format',
            metavar='<input-format>'
        )
        sub_commands_subparsers.required = True

        for sub_cmd in self._sub_commands_map.values():
            sub_cmd_parser = sub_commands_subparsers.add_parser(sub_cmd.name, help=sub_cmd.help_message)
            sub_cmd.build_parser(sub_cmd_parser)

    def parse_and_execute(self, parsed_args: Dict):
        self._sub_commands_map[parsed_args['input-format']].parse_and_execute(parsed_args)
        return 0



from abc import ABC, abstractmethod
from typing import Dict


class SubCommand(ABC):
    def __init__(self, name: str, help_message: str):
        self.name = name
        self.help_message = help_message

    @abstractmethod
    def build_parser(self, parser):
        """Take a vanilla argparser from main argparser's action sub argparser and add arguments

        :param parser: a vanilla argparser from main argparser's action sub argparser
        """

    @abstractmethod
    def parse_and_execute(self, parsed_args: Dict):
        """Use the parsed args from main argparser and execute the action

        :param parsed_args:
        """



"""Aito `Data API <https://aito.ai/docs/api/#database-api>`__ Request Class"""

import re
from abc import ABC, abstractmethod
from typing import Optional, Union, Dict, List

from aito.client import responses as aito_resp
from .aito_request import AitoRequest, _FinalRequest, _PatternEndpoint, _GetRequest, _PostRequest


class _DataAPIRequest(AitoRequest, ABC):
    """Request to manipulate the schema"""
    endpoint_prefix = f'{AitoRequest._api_version_endpoint_prefix}/data'

    @classmethod
    @abstractmethod
    def _check_method(cls, method: str) -> bool:
        return method in cls._request_methods

    @classmethod
    @abstractmethod
    def _check_endpoint(cls, endpoint: str) -> bool:
        return endpoint.startswith(cls.endpoint_prefix)

    @classmethod
    @abstractmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        for sub_cls in cls.__subclasses__():
            if sub_cls._check_method(method) and sub_cls._check_endpoint(endpoint):
                return sub_cls.make_request(method=method, endpoint=endpoint, query=query)
        raise ValueError(f"invalid {cls.__name__} with '{method}({endpoint})'")


class UploadEntriesRequest(_PostRequest, _PatternEndpoint, _DataAPIRequest):
    """Request to `Insert entries to a table <https://aito.ai/docs/api/#post-api-v1-data-table>`__"""
    response_cls = aito_resp.BaseResponse

    @classmethod
    def _endpoint_pattern(cls):
        return re.compile(f'^{cls.endpoint_prefix}/([^/".$\r\n\s]+)/batch$')

    def __init__(self, table_name: str, entries: List[Dict]):
        """

        :param table_name: the name of the table to be uploaded
        :type table_name: str
        :param entries: a list of the table entries
        :type entries: List[Dict]
        """
        endpoint = f'{self.endpoint_prefix}/{table_name}/batch'
        super().__init__(method=self.method, endpoint=endpoint, query=entries)

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        groups = cls._endpoint_to_captured_groups(endpoint=endpoint)
        return cls(table_name=groups(1), entries=query)


class DeleteEntriesRequest(_PostRequest, _FinalRequest, _DataAPIRequest):
    """Request to `Delete entries of a table <https://aito.ai/docs/api/#post-api-v1-data-delete>`__"""
    endpoint = f'{_DataAPIRequest.endpoint_prefix}/_delete'
    response_cls = aito_resp.BaseResponse


class InitiateFileUploadRequest(_PostRequest, _PatternEndpoint, _DataAPIRequest):
    """Request to `Initiate File Upload <https://aito.ai/docs/api/#post-api-v1-data-table-file>`__"""
    response_cls = aito_resp.BaseResponse

    @classmethod
    def _endpoint_pattern(cls):
        return re.compile(f'^{cls.endpoint_prefix}/([^/".$\r\n\s]+)/file$')

    def __init__(self, table_name: str):
        """

        :param table_name: the name of the table to be uploaded
        :type table_name: str
        """
        endpoint = f'{self.endpoint_prefix}/{table_name}/file'
        super().__init__(method=self.method, endpoint=endpoint)

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        groups = cls._endpoint_to_captured_groups(endpoint=endpoint)
        return cls(table_name=groups(1))


class TriggerFileProcessingRequest(_PostRequest, _PatternEndpoint, _DataAPIRequest):
    """Request to `Initiate File Upload <https://aito.ai/docs/api/#post-api-v1-data-table-file>`__"""
    response_cls = aito_resp.BaseResponse

    @classmethod
    def _endpoint_pattern(cls):
        return re.compile(f'^{cls.endpoint_prefix}/([^/".$\r\n\s]+)/file/(.+)$')

    def __init__(self, table_name: str, session_id: str):
        """

        :param table_name: the name of the table to be uploaded
        :type table_name: str
        :param session_id: The uuid of the file upload session from initiating file upload
        :type session_id: str
        """
        endpoint = f'{self.endpoint_prefix}/{table_name}/file/{session_id}'
        super().__init__(method=self.method, endpoint=endpoint)

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        groups = cls._endpoint_to_captured_groups(endpoint=endpoint)
        return cls(table_name=groups(1), session_id=groups(2))


class GetFileProcessingRequest(_GetRequest, TriggerFileProcessingRequest, _DataAPIRequest):
    """Request to `Initiate File Upload <https://aito.ai/docs/api/#post-api-v1-data-table-file>`__"""


"""

Data structure for the Aito Database Schema
"""

import itertools
import logging
from abc import abstractmethod, ABC
from collections import Counter
from csv import Sniffer, Error as csvError
from typing import List, Iterable, Dict, Optional
import re

import pandas as pd
from langdetect import detect_langs

from aito.utils._json_format import JsonFormat, JsonValidationError

LOG = logging.getLogger('AitoSchema')


class AitoSchema(JsonFormat, ABC):
    """The base class for Aito schema component

    """
    @property
    @abstractmethod
    def type(self):
        """the type of the schema component

        :rtype: str
        """
        pass

    @property
    @abstractmethod
    def comparison_properties(self) -> Iterable[str]:
        """properties of the schema object that will be used for comparison operation

        :rtype: Iterable[str]
        """
        pass

    def _compare_type(self, other) -> bool:
        try:
            return self.type == other.type
        except AttributeError:
            return False

    def _compare_properties(self, other) -> bool:
        return all([getattr(self, prop) == getattr(other, prop) for prop in self.comparison_properties])

    def __eq__(self, other):
        return self._compare_type(other) and self._compare_properties(other)


class AitoAnalyzerSchema(AitoSchema, ABC):
    """the base class for `Aito Analyzer <https://aito.ai/docs/api/#schema-analyzer>`__
    """
    _supported_analyzer_type = ['char-ngram', 'delimiter', 'language', 'token-ngram']
    _lang_detect_code_to_aito_code = {
        'ko': 'cjk',
        'ja': 'cjk',
        'zh-cn': 'cjk',
        'zh-tw': 'cjk',
        'pt': 'pt-br'
    }
    _supported_language_analyzer_iso_code_to_name = {
        'ar': 'arabic',
        'bg': 'bulgarian',
        'ca': 'catalan',
        'cjk': 'cjk',
        'cs': 'czech',
        'da': 'danish',
        'de': 'german',
        'el': 'greek',
        'en': 'english',
        'es': 'spanish',
        'eu': 'basque',
        'fa': 'persian',
        'fi': 'finnish',
        'fr': 'french',
        'ga': 'irish',
        'gl': 'galician',
        'hi': 'hindi',
        'hu': 'hungarian',
        'hy': 'armenian',
        'id': 'indonesian',
        'it': 'italian',
        'lv': 'latvian',
        'nl': 'dutch',
        'no': 'norwegian',
        'pt': 'portuguese',
        'pt - br': 'brazilian',
        'ro': 'romanian',
        'ru': 'russian',
        'sv': 'swedish',
        'th': 'thai',
        'tr': 'turkish'
    }
    _supported_language_analyzer_aliases = [
        al for iso_code_and_lang in _supported_language_analyzer_iso_code_to_name.items() for al in iso_code_and_lang
    ]
    _supported_analyzer_aliases = ['standard', 'whitespace'] + _supported_language_analyzer_aliases

    @property
    def type(self):
        return 'analyzer'

    @property
    @abstractmethod
    def analyzer_type(self) -> str:
        """the type of the analyzer

        :rtype: str
        """
        pass

    @classmethod
    @abstractmethod
    def json_schema(cls):
        return {
            'oneOf': [
                {'type': 'string'},
                {
                    'type': 'object',
                    'properties': {'type': {'enum': cls._supported_analyzer_type}},
                    'required': ['type']
                }
            ]
        }

    @classmethod
    @abstractmethod
    def from_deserialized_object(cls, obj):
        cls.json_schema_validate(obj)

        if isinstance(obj, str):
            return AitoAliasAnalyzerSchema.from_deserialized_object(obj)
        else:
            analyzer_type = obj.get('type')
            if analyzer_type == 'language':
                return AitoLanguageAnalyzerSchema.from_deserialized_object(obj)
            if analyzer_type == 'delimiter':
                return AitoDelimiterAnalyzerSchema.from_deserialized_object(obj)
            if analyzer_type == 'char-ngram':
                return AitoCharNGramAnalyzerSchema.from_deserialized_object(obj)
            if analyzer_type == 'token-ngram':
                return AitoTokenNgramAnalyzerSchema.from_deserialized_object(obj)

    def __eq__(self, other):
        self._compare_type(other)
        if self.analyzer_type != other.analyzer_type:
            if {self.analyzer_type, other.analyzer_type} == {'alias', 'language'}:
                alias_analyzer, language_analyzer = (self, other) if self.analyzer_type == 'alias' else (other, self)
                language_analyzer_args = [
                    getattr(language_analyzer, key)
                    for key in ('use_default_stop_words', 'custom_stop_words', 'custom_key_words')
                ]
                # if language_analyzer has the same language and use the default parameters
                if alias_analyzer.alias == language_analyzer.language and language_analyzer_args == [False, [], []]:
                    return True
            if {self.analyzer_type, other.analyzer_type} == {'alias', 'delimiter'}:
                alias_analyzer, delimiter_analyzer = (self, other) if self.analyzer_type == 'alias' else (other, self)
                if alias_analyzer.alias == 'whitespace' and delimiter_analyzer.delimiter == ' ' and \
                        delimiter_analyzer.trim_white_space:
                    return True
            return False
        return self._compare_properties(other)

    @staticmethod
    def _try_sniff_delimiter(sample: str, candidates='-,;:|\t') -> Optional[str]:
        try:
            return str(Sniffer().sniff(sample, candidates).delimiter)
        except csvError:
            # deprio whitespace in case tokens has trailing or leading whitespace
            try:
                return str(Sniffer().sniff(sample, ' ').delimiter)
            except csvError:
                return None

    @classmethod
    def _infer_delimiter(cls, samples: Iterable[str]) -> Optional[str]:
        """returns the most common inferred delimiter from the samples"""
        inferred_delimiter_counter = Counter([cls._try_sniff_delimiter(smpl) for smpl in samples])
        LOG.debug(f'inferred delimiter and count: {inferred_delimiter_counter}')
        most_common_delimiter_and_count = inferred_delimiter_counter.most_common(1)[0]
        return most_common_delimiter_and_count[0]

    @classmethod
    def _infer_language(cls, samples: Iterable[str]) -> Optional[str]:
        """infer language from samples"""
        concatenated_sample_text = ' '.join(samples)
        try:
            detected_langs_and_probs = detect_langs(concatenated_sample_text)
        except:
            detected_langs_and_probs = None
        if detected_langs_and_probs:
            LOG.debug(f'inferred languages and probabilities: {detected_langs_and_probs}')
            most_probable_lang_and_prob = detected_langs_and_probs[0]
            if most_probable_lang_and_prob.prob > 0.9:
                most_probable_lang = most_probable_lang_and_prob.lang
                if most_probable_lang in cls._lang_detect_code_to_aito_code:
                    most_probable_lang = cls._lang_detect_code_to_aito_code[most_probable_lang]
                if most_probable_lang in cls._supported_language_analyzer_iso_code_to_name:
                    return most_probable_lang
        return None

    @classmethod
    def infer_from_samples(cls, samples: Iterable[str], max_sample_size: int = 10000):
        """Infer an analyzer from the given samples

        :param samples: iterable of sample
        :type samples: Iterable
        :param max_sample_size: at most first max_sample_size will be used for inference, defaults to 10000
        :type max_sample_size: int
        :return: inferred Analyzer or None if no analyzer is applicable
        :rtype: Optional[AitoAnalyzerSchema]
        """
        sliced_samples = list(itertools.islice(samples, max_sample_size))
        detected_delimiter = cls._infer_delimiter(sliced_samples)
        if detected_delimiter is None or detected_delimiter.isalnum():
            return None
        if detected_delimiter == ' ':
            detected_language = cls._infer_language(sliced_samples)
            return AitoLanguageAnalyzerSchema(language=detected_language) if detected_language \
                else AitoAliasAnalyzerSchema(alias='whitespace')
        return AitoDelimiterAnalyzerSchema(delimiter=detected_delimiter)


class AitoAliasAnalyzerSchema(AitoAnalyzerSchema):
    """Aito `AliasAnalyzer <https://aito.ai/docs/api/#schema-alias-analyzer>`__ schema
    """
    def __init__(self, alias: str):
        """

        :param alias: the alias of the analyzer, standardize to the language name if the alias is a language ISO code
        :type alias: str
        """
        if not isinstance(alias, str):
            raise TypeError("alias must be of type str")
        alias = alias.lower().strip()
        if alias not in self._supported_analyzer_aliases:
            raise ValueError(f"unsupported alias {alias}")
        if alias in self._supported_language_analyzer_iso_code_to_name:
            alias = self._supported_language_analyzer_iso_code_to_name[alias]
        self._alias = alias

    @property
    def analyzer_type(self) -> str:
        return 'alias'

    @property
    def alias(self) -> str:
        """the alias of the analyzer

        :rtype: str
        """
        return self._alias

    @property
    def comparison_properties(self) -> Iterable[str]:
        return ['alias']

    @classmethod
    def json_schema(cls):
        return {'type': 'string', 'enum': AitoAnalyzerSchema._supported_analyzer_aliases}

    @classmethod
    def from_deserialized_object(cls, obj: str):
        cls.json_schema_validate(obj)
        return cls(alias=obj)

    def to_json_serializable(self) -> str:
        return self._alias


class AitoLanguageAnalyzerSchema(AitoAnalyzerSchema):
    """Aito `LanguageAnalyzer <https://aito.ai/docs/api/#schema-language-analyzer>`__ schema

    """
    def __init__(
            self,
            language: str,
            use_default_stop_words: bool = None,
            custom_stop_words: List[str] = None,
            custom_key_words: List[str] = None
    ):
        """

        :param language: the name or the ISO code of the language
        :type language: str
        :param use_default_stop_words: filter the language default stop words
        :type use_default_stop_words: bool, defaults to False
        :param custom_stop_words: words that will be filtered
        :type custom_stop_words: List[str], defaults to []
        :param custom_key_words: words that will not be featurized
        :type custom_key_words: List[str], defaults to []
        """
        self.language = language
        self.use_default_stop_words = use_default_stop_words
        self.custom_stop_words = custom_stop_words
        self.custom_key_words = custom_key_words if custom_key_words is not None else []

    @property
    def analyzer_type(self) -> str:
        return 'language'

    @property
    def language(self) -> str:
        """the language of the analyzer

        :rtype: str
        """
        return self._language

    @language.setter
    def language(self, value):
        self.json_schema_validate_with_schema(value, self.json_schema()['properties']['language'])
        if value in self._supported_language_analyzer_iso_code_to_name:
            value = self._supported_language_analyzer_iso_code_to_name[value]
        self._language = value

    @property
    def use_default_stop_words(self) -> bool:
        """filter the language default stop words

        :rtype: bool
        """
        return self._use_default_stop_words

    @use_default_stop_words.setter
    def use_default_stop_words(self, value):
        schema = self.json_schema()['properties']['useDefaultStopWords']
        if value is None:
            value = schema['default']
        else:
            self.json_schema_validate_with_schema(value, schema)
        self._use_default_stop_words = value

    @property
    def custom_stop_words(self) -> List[str]:
        """list of words that will be filtered

        :rtype: List[str]
        """
        return self._custom_stop_words

    @custom_stop_words.setter
    def custom_stop_words(self, value):
        schema = self.json_schema()['properties']['customStopWords']
        if value is None:
            value = schema['default']
        else:
            self.json_schema_validate_with_schema(value, schema)
        self._custom_stop_words = value

    @property
    def custom_key_words(self) -> List[str]:
        """list of words that will not be featurized

        :rtype: List[str]
        """
        return self._custom_key_words

    @custom_key_words.setter
    def custom_key_words(self, value):
        schema = self.json_schema()['properties']['customKeyWords']
        if value is None:
            value = schema['default']
        else:
            self.json_schema_validate_with_schema(value, schema)
        self._custom_key_words = value

    @property
    def comparison_properties(self) -> Iterable[str]:
        return ['language', 'use_default_stop_words', 'custom_stop_words', 'custom_key_words']

    @classmethod
    def json_schema(cls):
        return {
            'type': 'object',
            'properties': {
                'type': {'enum': ['language']},
                'language': {'type': 'string', 'enum': AitoAnalyzerSchema._supported_language_analyzer_aliases},
                'useDefaultStopWords': {'type': 'boolean', 'default': False},
                'customStopWords': {'type': 'array', 'items': {'type': 'string'}, 'default': []},
                'customKeyWords': {'type': 'array', 'items': {'type': 'string'}, 'default': []},
            },
            'required': ['type', 'language'],
            'additionalProperties': False
        }

    @classmethod
    def from_deserialized_object(cls, obj: Dict):
        cls.json_schema_validate(obj)
        return cls(
            language=obj.get('language'),
            use_default_stop_words=obj.get('useDefaultStopWords'),
            custom_stop_words=obj.get('customStopWords'),
            custom_key_words=obj.get('customKeyWords')
        )

    def to_json_serializable(self) -> Dict:
        return {
            'type': 'language',
            'language': self.language,
            'useDefaultStopWords': self.use_default_stop_words,
            'customStopWords': self.custom_stop_words,
            'customKeyWords': self.custom_key_words
        }


class AitoDelimiterAnalyzerSchema(AitoAnalyzerSchema):
    """Aito `DelimiterAnalyzer <https://aito.ai/docs/api/#schema-delimiter-analyzer>`__ schema

    """
    def __init__(self, delimiter: str, trim_white_space: bool = None):
        """

        :param delimiter: the delimiter
        :type delimiter: str
        :param trim_white_space: trim leading and trailing whitespaces of the features
        :type trim_white_space: bool, defaults to True
        """
        self.delimiter = delimiter
        self.trim_white_space = trim_white_space

    @property
    def analyzer_type(self) -> str:
        return 'delimiter'

    @property
    def delimiter(self):
        """the delimiter of the analyzer

        :rtype: str
        """
        return self._delimiter

    @delimiter.setter
    def delimiter(self, value):
        self.json_schema_validate_with_schema(value, self.json_schema()['properties']['delimiter'])
        self._delimiter = value

    @property
    def trim_white_space(self) -> bool:
        """trim leading and trailing whitespaces of the features

        :rtype: bool
        """
        return self._trim_white_space

    @trim_white_space.setter
    def trim_white_space(self, value):
        schema = self.json_schema()['properties']['trimWhitespace']
        if value is None:
            value = schema['default']
        else:
            self.json_schema_validate_with_schema(value, schema)
        self._trim_white_space = value

    @property
    def comparison_properties(self) -> Iterable[str]:
        return ['delimiter', 'trim_white_space']

    @classmethod
    def json_schema(cls):
        return {
            'type': 'object',
            'properties': {
                'type': {'enum': ['delimiter']},
                'delimiter': {'type': 'string'},
                'trimWhitespace': {'type': 'boolean', 'default': True}
            },
            'required': ['type', 'delimiter'],
            'additionalProperties': False
        }

    def to_json_serializable(self):
        return {
            'type': 'delimiter',
            'delimiter': self.delimiter,
            'trimWhitespace': self.trim_white_space
        }

    @classmethod
    def from_deserialized_object(cls, obj):
        cls.json_schema_validate(obj)
        return cls(
            delimiter=obj.get('delimiter'),
            trim_white_space=obj.get('trimWhitespace')
        )


class AitoCharNGramAnalyzerSchema(AitoAnalyzerSchema):
    """Aito `CharNGramAnalyzer <https://aito.ai/docs/api/#schema-char-n-gram-analyzer>`__ schema

    """
    def __init__(self, min_gram: int, max_gram: int):
        """

        :param min_gram: the minimum length of characters in a feature
        :type min_gram: int
        :param max_gram: the maximum length of characters in a feature
        :type max_gram: int
        """
        self.min_gram = min_gram
        self.max_gram = max_gram

    @property
    def analyzer_type(self) -> str:
        return 'char-ngram'

    @property
    def comparison_properties(self) -> Iterable[str]:
        return ['min_gram', 'max_gram']

    def to_json_serializable(self):
        return {
            'type': 'char-ngram',
            'minGram': self.min_gram,
            'maxGram': self.max_gram
        }

    @property
    def min_gram(self):
        return self._min_gram

    @min_gram.setter
    def min_gram(self, value):
        self.json_schema_validate_with_schema(value, self.json_schema()['properties']['minGram'])
        self._min_gram = value

    @property
    def max_gram(self):
        return self._max_gram

    @max_gram.setter
    def max_gram(self, value):
        self.json_schema_validate_with_schema(value, self.json_schema()['properties']['maxGram'])
        self._max_gram = value

    @classmethod
    def json_schema(cls):
        return {
            'type': 'object',
            'properties': {
                'type': {'enum': ['char-ngram']},
                'minGram': {'type': 'integer', 'minimum': 1},
                'maxGram': {'type': 'integer', 'minimum': 1}
            },
            'required': ['type', 'minGram', 'maxGram'],
            'additionalProperties': False
        }

    @classmethod
    def from_deserialized_object(cls, obj):
        cls.json_schema_validate(obj)
        return cls(
            min_gram=obj.get('minGram'),
            max_gram=obj.get('maxGram')
        )


class AitoTokenNgramAnalyzerSchema(AitoAnalyzerSchema):
    """Aito `TokenNGramAnalyzer <https://aito.ai/docs/api/#schema-token-n-gram-analyzer>`__ schema

    """
    def __init__(self, source: AitoAnalyzerSchema, min_gram: int, max_gram: int, token_separator: str = None):
        """

        :param source: the source analyzer to generate features before being combined into n-grams
        :type source: AitoAnalyzerSchema
        :param min_gram: the minimum length of characters in a feature
        :type min_gram: int
        :param max_gram: the maximum length of characters in a feature
        :type max_gram: int
        :param token_separator: the string used to join the features of the source analyzer
        :type token_separator: str, defaults to ' '
        """
        self.source = source
        self.min_gram = min_gram
        self.max_gram = max_gram
        self.token_separator = token_separator

    @property
    def analyzer_type(self) -> str:
        return 'token-ngram'

    @property
    def source(self):
        """the source analyzer

        :rtype: AitoAnalyzerSchema
        """
        return self._source

    @source.setter
    def source(self, value):
        if not isinstance(value, AitoAnalyzerSchema):
            raise TypeError('source analyzer must be of type AitoAnalyzerSchema')
        self._source = value

    @property
    def min_gram(self):
        return self._min_gram

    @min_gram.setter
    def min_gram(self, value):
        self.json_schema_validate_with_schema(value, self.json_schema()['properties']['minGram'])
        self._min_gram = value

    @property
    def max_gram(self):
        return self._max_gram

    @max_gram.setter
    def max_gram(self, value):
        self.json_schema_validate_with_schema(value, self.json_schema()['properties']['maxGram'])
        self._max_gram = value

    @property
    def token_separator(self) -> str:
        """the string that will be used to join the features

        :rtype: str
        """
        return self._token_separator

    @token_separator.setter
    def token_separator(self, value):
        schema = self.json_schema()['properties']['tokenSeparator']
        if value is None:
            value = schema['default']
        else:
            self.json_schema_validate_with_schema(value, schema)
        self._token_separator = value

    @property
    def comparison_properties(self) -> Iterable[str]:
        return ['source', 'min_gram', 'max_gram', 'token_separator']

    @classmethod
    def json_schema(cls):
        return {
            'type': 'object',
            'properties': {
                'type': {'enum': ['token-ngram']},
                'source': {'oneOf': [{'type': 'string'}, {'type': 'object'}]},
                'minGram': {'type': 'integer', 'minimum': 1},
                'maxGram': {'type': 'integer', 'minimum': 1},
                'tokenSeparator': {'type': 'string', 'default': ' '}
            },
            'required': ['type', 'source', 'minGram', 'maxGram'],
            'additionalProperties': False
        }

    def to_json_serializable(self):
        return {
            'type': 'token-ngram',
            'source': self.source.to_json_serializable(),
            'minGram': self.min_gram,
            'maxGram': self.max_gram,
            'tokenSeparator': self.token_separator
        }

    @classmethod
    def from_deserialized_object(cls, obj):
        cls.json_schema_validate(obj)
        return cls(
            source=AitoAnalyzerSchema.from_deserialized_object(obj.get('source')),
            min_gram=obj.get('minGram'),
            max_gram=obj.get('maxGram'),
            token_separator=obj.get('tokenSeparator')
        )

class DataSeriesProperties :
    """DataSeriesProperties is an utility class that is used to inspect pandas data series
       properties and infer the Aito data type based on it. It checks the maximum and
       minimum value and it will convert integers into string, if Aito does support the
       numeric range.
    """

    _pandas_dtypes_name_to_aito_type = {
        'string': 'Text',
        'bytes': 'String',
        'floating': 'Decimal',
        'integer': 'Int',
        'mixed-integer': 'Text',
        'mixed-integer-float': 'Decimal',
        'decimal': 'Decimal',
        'complex': 'Decimal',
        'categorical': 'String',
        'boolean': 'Boolean',
        'datetime64': 'String',
        'datetime': 'String',
        'date': 'String',
        'timedelta64': 'String',
        'timedelta': 'String',
        'time': 'String',
        'period': 'String',
        'mixed': 'Text',
        'empty': 'String'
    }

    MAX_INT_VALUE = 2147483647
    MIN_INT_VALUE = -2147483648

    def __init__(self, pandas_dtype: str, min_value, max_value):
        self.pandas_dtype = pandas_dtype
        self.min_value = min_value
        self.max_value = max_value

        if pandas_dtype == 'integer' and (max_value > self.MAX_INT_VALUE or min_value < self.MIN_INT_VALUE):
            self.target_aito_dtype = 'String' # neither pandas (!) or Aito supports integers this large
        else:
            self.target_aito_dtype = DataSeriesProperties.pandas_dtype_to_aito_dtype(pandas_dtype)


    @classmethod
    def pandas_dtype_to_aito_dtype(cls, pandas_dtype) -> str:
        """ Converts a pandas data type into Aito data type

        :rtype: str
        """
        return cls._pandas_dtypes_name_to_aito_type[pandas_dtype]

    def to_data_type_schema(self) -> 'AitoDataTypeSchema':
        """ Provides the AitoDataTypeSchema that is inferred for this data series
        """
        return AitoDataTypeSchema.from_deserialized_object(self.target_aito_dtype)

    @classmethod
    def _infer_from_pandas_series(cls, series: pd.Series, max_sample_size: int = 100000) -> 'DataSeriesProperties':
        """Infer aito column type from a Pandas Series

        :param series: input Pandas Series
        :type series: pd.Series
        :param max_sample_size: maximum sample size that will be used for type inference, defaults to 100000
        :type max_sample_size: int, optional
        :raises Exception: fail to infer type
        :return: inferred Aito type
        :rtype: str
        """
        sampled_values = series.values if len(series) < max_sample_size else series.sample(max_sample_size).values

        if len(series) == series.isna().sum(): # pandas will infer empty series as floating point as default
            inferred_dtype = 'empty'
        else:
            inferred_dtype = pd.api.types.infer_dtype(sampled_values)

        lower_bound = None
        upper_bound = None
        if inferred_dtype == 'integer':
            lower_bound = sampled_values.min()
            upper_bound = sampled_values.max()
            # See integer MAX_VALUE and MIN_VALUE in https://docs.oracle.com/javase/8/docs/api/constant-values.html
            LOG.debug(f'inferred pandas dtype: {inferred_dtype}')
        if inferred_dtype not in cls._pandas_dtypes_name_to_aito_type:
            LOG.debug(f'failed to convert pandas dtype {inferred_dtype} to aito dtype')
            raise Exception(f'failed to infer aito data type')

        return DataSeriesProperties(inferred_dtype, lower_bound, upper_bound)


class AitoDataTypeSchema(AitoSchema, ABC):
    """The base class for Aito DataType"""

    _supported_data_types = ["Boolean", "Decimal", "Int", "String", "Text"]

    def __init__(self, aito_dtype: str):
        """

        :param aito_dtype: the Aito data type
        :type aito_dtype: str
        """
        if aito_dtype not in self._supported_data_types:
            raise TypeError(
                f"unrecognized data type `{aito_dtype}`. "
                f"Data type must be one of {'|'.join(self._supported_data_types)}"
            )
        self._aito_dtype = aito_dtype

    @property
    def type(self):
        return 'dtype'

    @property
    def aito_dtype(self) -> str:
        """the data type

        :rtype: str
        """
        return self._aito_dtype

    @property
    def is_string(self) -> bool:
        """returns True if the data type is String

        :rtype: bool
        """
        return self._aito_dtype == 'String'

    @property
    def is_text(self) -> bool:
        """returns True if the data type is Text

        :rtype: bool
        """
        return self._aito_dtype == 'Text'

    @property
    def is_bool(self) -> bool:
        """returns True if the data type is Boolean

        :rtype: bool
        """
        return self._aito_dtype == 'Boolean'

    @property
    def is_int(self) -> bool:
        """returns True if the data type is Int

        :rtype: bool
        """
        return self._aito_dtype == 'Int'

    @property
    def is_decimal(self) -> bool:
        """returns True if the data type is Decimal

        :rtype: bool
        """
        return self._aito_dtype == 'Decimal'

    def to_json_serializable(self) -> str:
        return self._aito_dtype

    @abstractmethod
    def to_python_type(self):
        """the equivalent python type
        """
        pass

    @classmethod
    def json_schema(cls):
        return {'type': 'string', 'enum': cls._supported_data_types}

    @classmethod
    def from_deserialized_object(cls, obj: str) -> 'AitoDataTypeSchema':
        cls.json_schema_validate(obj)
        if obj == 'Boolean':
            return AitoBooleanType()
        if obj == 'Int':
            return AitoIntType()
        if obj == 'Decimal':
            return AitoDecimalType()
        if obj == 'String':
            return AitoStringType()
        if obj == 'Text':
            return AitoTextType()

    @property
    def comparison_properties(self) -> Iterable[str]:
        return ['aito_dtype']

    @classmethod
    def _infer_from_pandas_series(cls, series: pd.Series, max_sample_size: int = 100000) -> 'AitoDataTypeSchema':
        """Infer aito column type from a Pandas Series

        :param series: input Pandas Series
        :type series: pd.Series
        :param max_sample_size: maximum sample size that will be used for type inference, defaults to 100000
        :type max_sample_size: int, optional
        :raises Exception: fail to infer type
        :return: inferred Aito type
        :rtype: str
        """
        return DataSeriesProperties._infer_from_pandas_series(series, max_sample_size).to_data_type_schema()

    @classmethod
    def infer_from_samples(cls, samples: Iterable, max_sample_size: int = 100000) -> 'AitoDataTypeSchema':
        """infer AitoDataType from the given samples

        :param samples: iterable of sample
        :type samples: Iterable
        :param max_sample_size: at most first max_sample_size will be used for inference, defaults to 100000
        :type max_sample_size: int
        :return: inferred Aito column type
        :rtype: str
        """
        try:
            casted_samples = pd.Series(itertools.islice(samples, max_sample_size))
        except Exception as e:
            LOG.debug(f'failed to cast samples ({list(itertools.islice(samples, 10))}, ...)  to pandas Series: {e}')
            raise Exception(f'failed to infer aito type')
        return cls._infer_from_pandas_series(casted_samples)


class AitoBooleanType(AitoDataTypeSchema):
    """Aito `Boolean Type <https://aito.ai/docs/api/#schema-boolean-type>`__"""
    def __init__(self):
        super().__init__('Boolean')

    def to_python_type(self):
        return bool


class AitoIntType(AitoDataTypeSchema):
    """Aito `Int Type <https://aito.ai/docs/api/#schema-int-type>`__"""
    def __init__(self):
        super().__init__('Int')

    def to_python_type(self):
        return int


class AitoDecimalType(AitoDataTypeSchema):
    """Aito `Decimal Type <https://aito.ai/docs/api/#schema-decimal-type>`__"""
    def __init__(self):
        super().__init__('Decimal')

    def to_python_type(self):
        return float


class AitoStringType(AitoDataTypeSchema):
    """Aito `String Type <https://aito.ai/docs/api/#schema-string-type>`__"""
    def __init__(self):
        super().__init__('String')

    def to_python_type(self):
        return str

class AitoTextType(AitoDataTypeSchema):
    """Aito `Text Type <https://aito.ai/docs/api/#schema-text-type>`__"""
    def __init__(self):
        super().__init__('Text')

    def to_python_type(self):
        return str


class AitoColumnLinkSchema(AitoSchema):
    """Link to a column of another table"""

    # Essentially a combination of the table- and column-regexes
    _column_link_regex = r'^[^\/"\.$\r\n\s]+\.[^\/"\.$\r\n\s]+$'

    def __init__(self, table_name: str, column_name: str):
        """

        :param table_name: the name of the linked table
        :type table_name: str
        :param column_name: the name of the linked column
        :type column_name: str
        """
        if not bool(re.match(AitoDatabaseSchema._table_name_regex, table_name)):
            raise ValueError(f'Linked table name in "{table_name}.{column_name}" is not valid')
        if not bool(re.match(AitoTableSchema._column_name_regex, column_name)):
            raise ValueError(f'Linked column name in "{table_name}.{column_name}" is not valid')

        self._table_name = table_name
        self._column_name = column_name

    @property
    def type(self):
        return 'columnLink'

    @property
    def table_name(self) -> str:
        """the name of the linked table

        :rtype: str
        """
        return self._table_name

    @property
    def column_name(self) -> str:
        """the name of the linked column

        :rtype: str
        """
        return self._column_name

    def to_json_serializable(self):
        return f"{self.table_name}.{self.column_name}"

    @classmethod
    def json_schema(cls):
        return {'type': 'string', 'pattern': AitoColumnLinkSchema._column_link_regex}

    @classmethod
    def from_deserialized_object(cls, obj: str) -> "AitoColumnLinkSchema":
        cls.json_schema_validate(obj)
        parts = obj.split('.')
        return cls(parts[0], parts[1])

    @property
    def comparison_properties(self) -> Iterable[str]:
        return ['table_name', 'column_name']


class AitoColumnTypeSchema(AitoSchema):
    """Aito `ColumnType <https://aito.ai/docs/api/#schema-column-type>`__ schema
    """
    def __init__(
            self,
            data_type: AitoDataTypeSchema,
            nullable: bool = None,
            link: AitoColumnLinkSchema = None,
            analyzer: AitoAnalyzerSchema = None
    ):
        """

        :param data_type: the type of the column
        :type data_type: AitoDataTypeSchema
        :param nullable: when true, `null` values are allowed
        :type nullable: bool, default to False
        :param link: path to a linked column
        :type link: AitoColumnLink, optional
        :param analyzer: the analyzer of the column if the column is of type Text
        :type analyzer: AnalyzerSchema, optional
        """
        self._analyzer = None  # FIXME: Hacky way to avoid data type dependency on analyzer
        self.data_type = data_type
        self.nullable = nullable
        self.link = link
        self.analyzer = analyzer

    @property
    def type(self):
        return 'column'

    @property
    def data_type(self):
        """The data type of the column

        :rtype: AitoDataTypeSchema
        """
        return self._data_type

    @data_type.setter
    def data_type(self, value):
        if not isinstance(value, AitoDataTypeSchema):
            raise TypeError('data_type must be of type AitoDataTypeSchema')
        if self.analyzer and not value.is_text:
            raise ValueError(f"{value} does not support analyzer")
        self._data_type = value

    @property
    def nullable(self) -> bool:
        """returns True if the column allow `null` value

        :rtype: bool
        """
        return self._nullable

    @nullable.setter
    def nullable(self, value):
        schema = self.json_schema()['properties']['nullable']
        if value is None:
            value = schema['default']
        else:
            self.json_schema_validate_with_schema(value, schema)
        self._nullable = value

    def to_conversion(self):
        if self._nullable:
            tpt = self._data_type.to_python_type()
            return lambda x: None if pd.isna(x) else tpt(x)
        else:
            return self._data_type.to_python_type()

    @property
    def link(self):
        """the link of the column

        :rtype: AitoColumnLinkSchema
        """
        return self._link

    @link.setter
    def link(self, value):
        if value is not None:
            if not isinstance(value, AitoColumnLinkSchema):
                raise TypeError('link must be of type AitoColumnLinkSchema')
        self._link = value

    @property
    def analyzer(self):
        """the analyzer of the column

        :rtype: AitoAnalyzerSchema
        """
        return self._analyzer

    @analyzer.setter
    def analyzer(self, value):
        if value is not None:
            if not self.data_type.is_text:
                raise ValueError(f"{self.data_type} does not support analyzer")
            if not isinstance(value, AitoAnalyzerSchema):
                raise TypeError('analyzer must be of type AitoAnalyzerSchema')
        self._analyzer = value

    @property
    def has_link(self) -> bool:
        """return true if the column is linked to another column"""
        return self.link is not None

    @property
    def comparison_properties(self) -> Iterable[str]:
        return ['data_type', 'nullable', 'link', 'analyzer']

    def to_json_serializable(self):
        data = {
            'type': self.data_type.to_json_serializable(),
            'nullable': self.nullable,
            'link': self.link.to_json_serializable() if self.link is not None else None,
            'analyzer': self.analyzer.to_json_serializable() if self.analyzer else self.analyzer
        }
        return {
            key: val for key, val in data.items() if val is not None
        }

    @classmethod
    def json_schema(cls):
        return {
            'type': 'object',
            'properties': {
                'type': {},
                'nullable': {'type': 'boolean', 'default': False},
                'link': {},
                'analyzer': {}
            },
            'required': ['type'],
            'additionalProperties': False
        }

    @classmethod
    def from_deserialized_object(cls, obj: Dict):
        cls.json_schema_validate(obj)
        data_type = AitoDataTypeSchema.from_deserialized_object(obj.get('type'))
        analyzer_data = obj.get('analyzer')
        analyzer = AitoAnalyzerSchema.from_deserialized_object(analyzer_data) if analyzer_data is not None else None
        link_data = obj.get('link')
        link = AitoColumnLinkSchema.from_deserialized_object(link_data) if link_data is not None else None
        return cls(data_type=data_type, nullable=obj.get('nullable'), link=link, analyzer=analyzer)

    @classmethod
    def _infer_from_pandas_series(cls, series: pd.Series, max_sample_size: int = 100000) -> 'AitoColumnTypeSchema':
        samples = series if len(series) < max_sample_size else series.sample(max_sample_size)

        col_nullable = True if series.isna().sum() > 0 else False
        col_analyzer = None
        col_aito_type = AitoDataTypeSchema._infer_from_pandas_series(samples, max_sample_size)

        if col_aito_type.is_text:
            col_analyzer = AitoAnalyzerSchema.infer_from_samples(samples.dropna().__iter__())
            if not col_analyzer:
                col_aito_type = AitoStringType()
        return cls(
            data_type=col_aito_type, nullable=col_nullable, analyzer=col_analyzer
        )

    @classmethod
    def infer_from_samples(cls, samples: Iterable, max_sample_size: int = 100000) -> 'AitoColumnTypeSchema':
        """infer AitoColumnType from the given samples

        :param samples: iterable of sample
        :type samples: Iterable
        :param max_sample_size: at most first max_sample_size will be used for inference, defaults to 100000
        :type max_sample_size: int
        :return: inferred Aito column type
        :rtype: str
        """
        try:
            casted_samples = pd.Series(itertools.islice(samples, max_sample_size))
        except Exception as e:
            LOG.debug(f'failed to cast samples ({list(itertools.islice(samples, 10))}, ...)  to pandas Series: {e}')
            raise Exception(f'failed to infer column type')
        return cls._infer_from_pandas_series(casted_samples)


class AitoTableSchema(AitoSchema):
    """Aito Table schema contains the columns and their schema

    Can be thought of as a dict-like container for :class:`.AitoColumnTypeSchema` objects

    Infer AitoTableSchema from a Pandas DataFrame

    >>> import pandas as pd
    >>> df = pd.DataFrame(data={'id': [1, 2], 'name': ['Neil', 'Buzz']})
    >>> table_schema = AitoTableSchema.infer_from_pandas_data_frame(df)
    >>> print(table_schema.to_json_string(indent=2, sort_keys=True))
    {
      "columns": {
        "id": {
          "nullable": false,
          "type": "Int"
        },
        "name": {
          "nullable": false,
          "type": "String"
        }
      },
      "type": "table"
    }

    >>> print(table_schema['name'])
    {"type": "String", "nullable": false}

    change the property of a column

    >>> table_schema['name'].nullable = True
    >>> print(table_schema['name'])
    {"type": "String", "nullable": true}

    add a column to the table schema

    >>> table_schema['description'] = AitoColumnTypeSchema(AitoTextType(), nullable=True)
    >>> print(table_schema.to_json_string(indent=2, sort_keys=True))
    {
      "columns": {
        "description": {
          "nullable": true,
          "type": "Text"
        },
        "id": {
          "nullable": false,
          "type": "Int"
        },
        "name": {
          "nullable": true,
          "type": "String"
        }
      },
      "type": "table"
    }

    delete a column in the table schema

    >>> del table_schema['description']
    >>> table_schema.columns
    ['id', 'name']

    check if a column exist in the table schema

    >>> 'id' in table_schema
    True

    iterate over the table schema
    >>> for col in table_schema:
    ...     table_schema[col].nullable = False
    """

    _column_name_regex = r'^[^\/".$\r\n\s]+$'

    def __init__(self, columns: Dict[str, AitoColumnTypeSchema]):
        """

        :param columns: a dictionary of the table's columns' name and schema
        :type columns: Dict[str, AitoColumnTypeSchema]
        """
        invalid_columns = []
        for key in columns.keys():
            if not bool(re.match(AitoTableSchema._column_name_regex, key)):
                invalid_columns.append(key)

        if len(invalid_columns) > 0:
            raise ValueError(f'Column names are not valid: {invalid_columns}')

        self._columns = columns

    @property
    def type(self):
        return 'table'

    @property
    def comparison_properties(self) -> Iterable[str]:
        return ['columns_schemas']

    @property
    def columns(self) -> List[str]:
        """list of the table's columns name

        :rtype: List[str]
        """
        return sorted(list(self._columns.keys()))

    @property
    def columns_schemas(self) -> Dict[str, AitoColumnTypeSchema]:
        """a dictionary contains the names of the table columns and its corresponding schemas

        :rtype: Dict[str, AitoColumnTypeSchema]
        """
        return self._columns

    @property
    def links(self) -> Optional[Dict[str, AitoColumnLinkSchema]]:
        """a dictionary contains the names of the table columns and its corresponding link

        :rtype: Dict[str, AitoColumnLinkSchema]
        """
        return {col_name: col_schema.link for col_name, col_schema in self._columns.items() if col_schema.has_link}

    def __getitem__(self, column_name: str):
        """get the column schema of the specified column name
        """
        if not isinstance(column_name, str):
            raise TypeError('the name of the column must be of type string')
        if column_name not in self._columns:
            raise KeyError(f'column `{column_name}` does not exist')
        return self._columns[column_name]

    def __setitem__(self, column_name: str, value):
        """update the column schema of the specified column name
        """
        if not isinstance(column_name, str):
            raise TypeError('the name of the column must be of type string')

        if not bool(re.match(AitoTableSchema._column_name_regex, column_name)):
            raise ValueError(f'The column name {column_name} is not valid')

        if not isinstance(value, AitoColumnTypeSchema):
            raise TypeError('column schema must be of type AitoColumnTypeSchema')
        self._columns[column_name] = value

    def __delitem__(self, key):
        """delete a column
        """
        if key not in self._columns:
            raise KeyError(f'column `{key}` does not exist')
        del self._columns[key]

    def __contains__(self, item):
        return item in self._columns

    def __iter__(self):
        return iter(self._columns)

    def has_column(self, column_name: str) -> bool:
        """check if the table has the specified column

        :param column_name: the name of the column
        :type column_name: str
        :return: true if the table has the specified column
        :rtype: bool
        """
        return column_name in self._columns

    @classmethod
    def json_schema(cls):
        pattern = AitoTableSchema._column_name_regex
        return {
            'type': 'object',
            'properties': {
                'type': {'const': 'table'},
                'columns': {
                    'type': 'object',
                    'minProperties': 1,
                    'propertyNames': {'pattern': pattern}
                }
            },
            'required': ['type', 'columns'],
            'additionalProperties': False
        }

    def to_json_serializable(self):
        return {
            "type": "table",
            "columns": {
                col_name: col_type.to_json_serializable() for col_name, col_type in self._columns.items()
            }
        }

    @classmethod
    def from_deserialized_object(cls, obj):
        cls.json_schema_validate(obj)

        columns_data = obj.get('columns')
        columns = {
            col_name: AitoColumnTypeSchema.from_deserialized_object(col_data)
            for col_name, col_data in columns_data.items()
        }
        return cls(columns=columns)

    @classmethod
    def infer_from_pandas_data_frame(cls, df: pd.DataFrame, max_sample_size: int = 100000) -> 'AitoTableSchema':
        """Infer a TableSchema from a Pandas DataFrame

        :param df: input Pandas DataFrame
        :type df: pd.DataFrame
        :param max_sample_size: maximum number of rows that will be used for inference, defaults to 100000
        :type max_sample_size: int, optional
        :raises Exception: an error occurred during column type inference
        :return: inferred table schema
        :rtype: Dict
        """
        LOG.debug('inferring table schema...')
        columns_schema = {}
        for col in df.columns.values:
            LOG.debug(f'inferring column {col}...')
            col_type_schema = AitoColumnTypeSchema._infer_from_pandas_series(df[col], max_sample_size)
            columns_schema[col] = col_type_schema
        table_schema = cls(columns=columns_schema)
        LOG.debug('inferred table schema')
        return table_schema


class AitoDatabaseSchema(AitoSchema):
    """Aito Database Schema

    Can be thought of as a dict-like container for :class:`.AitoTableSchema` objects
    """

    _table_name_regex = r'^[^\/".$\r\n\s]+$'

    def __init__(self, tables: Dict[str, AitoTableSchema]):
        invalid_tables = []
        for key in tables.keys():
            if not bool(re.match(AitoDatabaseSchema._table_name_regex, key)):
                invalid_tables.append(key)

        if len(invalid_tables) > 0:
            raise ValueError(f'Column names are not valid: {invalid_tables}')

        self._tables = tables

    def _validate_links(self):
        for tbl_name, tbl_schema in self._tables.items():
            for col_name, col_schema in tbl_schema.columns_schemas.items():
                if col_schema.has_link:
                    link = col_schema.link
                    linked_tbl_name = link.table_name
                    if not self.has_table(linked_tbl_name):
                        raise ValueError(f'column {tbl_name}.{col_name} link to non-exist table {linked_tbl_name}')

                    linked_table_schema = self._tables[linked_tbl_name]
                    linked_col_name = link.column_name
                    if linked_col_name not in linked_table_schema:
                        raise ValueError(f'column {tbl_name}.{col_name} link to non-exist column '
                                         f'{linked_tbl_name}.{linked_col_name}')

    @property
    def type(self):
        return 'database'

    @property
    def tables(self) -> List[str]:
        """a list contains the names of the database's tables

        :rtype:
        """
        return list(self._tables.keys())

    @property
    def tables_schemas(self) -> Dict[str, AitoTableSchema]:
        """a dictionary contains the name of the database's tables and its corresponding schemas

        :rtype: Dict[str, AitoTableSchema]
        """
        return self._tables

    @property
    def comparison_properties(self) -> Iterable[str]:
        return ['tables_schemas']

    def to_json_serializable(self):
        return {
            "schema": {
                tbl_name: tbl_schema.to_json_serializable() for tbl_name, tbl_schema in self._tables.items()
            }
        }

    def __getitem__(self, table_name: str) -> AitoTableSchema:
        """get the schema of the specified table
        """
        if not isinstance(table_name, str):
            raise TypeError('the name of the column must be of type string')
        if table_name not in self._tables:
            raise KeyError(f'table `{table_name}` does not exist')
        return self._tables[table_name]

    def __setitem__(self, table_name: str, value):
        """update the schema of the specified table
        """
        if not isinstance(table_name, str):
            raise TypeError('the name of the table must be of type string')

        if not bool(re.match(AitoDatabaseSchema._table_name_regex, table_name)):
            raise ValueError(f'Table name {table_name} is not valid')

        if not isinstance(value, AitoTableSchema):
            raise TypeError('the table schema must be of type AitoTableSchema')
        self._tables[table_name] = value

    def __delitem__(self, key):
        """delete a column
        """
        if key not in self._tables:
            raise KeyError(f'table `{key}` does not exist')
        del self._tables[key]

    def __contains__(self, item):
        return item in self._tables

    def __iter__(self):
        return iter(self._tables)

    def has_table(self, table_name: str) -> bool:
        """check if the database has the specified table

        :param table_name: the name of the column
        :type table_name: str
        :return: true if the table has the specified column
        :rtype: bool
        """
        return table_name in self._tables

    @classmethod
    def json_schema(cls):
        return {
            'type': 'object',
            'properties': {
                'schema': {
                    'type': 'object',
                    'propertyNames': {'pattern': AitoDatabaseSchema._table_name_regex}
                }
            },
            'required': ['schema'],
            'additionalProperties': False
        }

    @classmethod
    def from_deserialized_object(cls, obj):
        cls.json_schema_validate(obj)
        schema_data = obj.get('schema')
        tables = {
            tbl_name: AitoTableSchema.from_deserialized_object(tbl_data) for tbl_name, tbl_data in schema_data.items()
        }
        return cls(tables=tables)

    def reachable_columns(self, table_name) -> List[str]:
        """return the name of the columns that can be reached from the specified table if the table has link,
        including the columns of the table

        :param table_name: the name of the table
        :type table_name: str
        :return: list of linked columns
        :rtype: List[str]
        """
        table_schema = self.__getitem__(table_name)
        linked_columns = table_schema.columns
        for col_name, col_schema in table_schema.columns_schemas.items():
            if col_schema.has_link:
                linked_table_name = col_schema.link.table_name
                linked_table_schema = self.__getitem__(linked_table_name)
                linked_columns += [
                    f'{linked_table_name}.{linked_tbl_col}' for linked_tbl_col in linked_table_schema.columns
                ]
        return linked_columns
